# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'about_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.resize(594, 379)
        self.verticalLayout = QtWidgets.QVBoxLayout(AboutDialog)
        self.verticalLayout.setContentsMargins(6, 6, 6, 6)
        self.verticalLayout.setSpacing(3)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setContentsMargins(-1, -1, 16, -1)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_2 = QtWidgets.QLabel(AboutDialog)
        self.label_2.setText("")
        self.label_2.setPixmap(QtGui.QPixmap(":/main_gui/logo"))
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label = QtWidgets.QLabel(AboutDialog)
        font = QtGui.QFont()
        font.setPointSize(14)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.verticalLayout_3.addWidget(self.label)
        self.label_3 = QtWidgets.QLabel(AboutDialog)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setItalic(True)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.contentBoxLayout = QtWidgets.QVBoxLayout()
        self.contentBoxLayout.setContentsMargins(0, 20, -1, -1)
        self.contentBoxLayout.setObjectName("contentBoxLayout")
        self.contentLabel = QtWidgets.QLabel(AboutDialog)
        font = QtGui.QFont()
        font.setPointSize(11)
        self.contentLabel.setFont(font)
        self.contentLabel.setTextFormat(QtCore.Qt.RichText)
        self.contentLabel.setWordWrap(True)
        self.contentLabel.setObjectName("contentLabel")
        self.contentBoxLayout.addWidget(self.contentLabel)
        self.verticalLayout_3.addLayout(self.contentBoxLayout)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem1)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(AboutDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(AboutDialog)
        self.buttonBox.accepted.connect(AboutDialog.accept)
        self.buttonBox.rejected.connect(AboutDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(AboutDialog)

    def retranslateUi(self, AboutDialog):
        _translate = QtCore.QCoreApplication.translate
        AboutDialog.setWindowTitle(_translate("AboutDialog", "About MDT"))
        self.label.setText(_translate("AboutDialog", "MDT"))
        self.label_3.setText(_translate("AboutDialog", "Microstructure Diffusion Toolbox"))
        self.contentLabel.setText(_translate("AboutDialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Droid Sans\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:11pt;\">Version: {version}</span></p>\n"
"<p style=\" margin-top:20px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:11pt;\">The Microstructure Diffusion Toolbox is a model recovery toolbox primarily meant for diffusion MRI analysis.</span></p>\n"
"<p style=\" margin-top:20px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:11pt;\">Software development by Robbert Harms, under the (Phd) supervision of Alard Roebroeck, at Maastricht University.</span></p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:11pt;\">Contributors:</span></p>\n"
"<ul style=\"margin-top: 0px; margin-bottom: 0px; margin-left: 0px; margin-right: 0px; -qt-list-indent: 1;\"><li style=\" font-size:11pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Robbert Harms</li>\n"
"<li style=\" font-size:11pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Alard Roebroeck</li>\n"
"<li style=\" font-size:11pt;\" style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Francisco Fritz</li></ul></body></html>"))

from . import main_gui_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'dialog_get_example_data.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_GetExampleDataDialog(object):
    def setupUi(self, GetExampleDataDialog):
        GetExampleDataDialog.setObjectName("GetExampleDataDialog")
        GetExampleDataDialog.resize(691, 208)
        self.verticalLayout = QtWidgets.QVBoxLayout(GetExampleDataDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(GetExampleDataDialog)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(GetExampleDataDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_3.addWidget(self.label_4)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.line = QtWidgets.QFrame(GetExampleDataDialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.label = QtWidgets.QLabel(GetExampleDataDialog)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_7 = QtWidgets.QLabel(GetExampleDataDialog)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 0, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.outputFile = QtWidgets.QLineEdit(GetExampleDataDialog)
        self.outputFile.setObjectName("outputFile")
        self.horizontalLayout_5.addWidget(self.outputFile)
        self.outputFileSelect = QtWidgets.QPushButton(GetExampleDataDialog)
        self.outputFileSelect.setObjectName("outputFileSelect")
        self.horizontalLayout_5.addWidget(self.outputFileSelect)
        self.gridLayout.addLayout(self.horizontalLayout_5, 0, 1, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.line_3 = QtWidgets.QFrame(GetExampleDataDialog)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.buttonBox = QtWidgets.QDialogButtonBox(GetExampleDataDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(GetExampleDataDialog)
        self.buttonBox.accepted.connect(GetExampleDataDialog.accept)
        self.buttonBox.rejected.connect(GetExampleDataDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(GetExampleDataDialog)
        GetExampleDataDialog.setTabOrder(self.outputFile, self.outputFileSelect)

    def retranslateUi(self, GetExampleDataDialog):
        _translate = QtCore.QCoreApplication.translate
        GetExampleDataDialog.setWindowTitle(_translate("GetExampleDataDialog", "Get example data"))
        self.label_3.setText(_translate("GetExampleDataDialog", "Get example data"))
        self.label_4.setText(_translate("GetExampleDataDialog", "Loads the MDT example data from the installation files"))
        self.label.setText(_translate("GetExampleDataDialog", "This will write the MDT example data (b1k_b2k and b6k datasets) to the indicated directory. You can use this data for testing MDT on your computer. These example datasets are contained within the MDT package and as such are available on every machine with MDT installed."))
        self.label_7.setText(_translate("GetExampleDataDialog", "Output folder:"))
        self.outputFileSelect.setText(_translate("GetExampleDataDialog", "File browser"))




#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Evaluate an expression on a protocol.

This is meant to quickly change a protocol using mathematical expressions. The expressions
can be any valid python string separated if needed with the semicolon (;).

The columns of the input protocol are loaded and stored as arrays with as variable names the names of the
columns. Next, the expression is evaluated on those columns and the result is stored in the indicated file.

Columns can easily be removed with the python 'del' command. New columns can easily be added by assignment.
When adding a column, the value can either be a scalar or a vector.

Additionally, the numpy library is available with prefix 'np.'.
"""
import argparse
import os
import numpy as np
import mdt
from argcomplete.completers import FilesCompleter
import textwrap

from mdt.protocols import Protocol
from mdt.lib.shell_utils import BasicShellApplication
from mot.lib.utils import is_scalar

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MathProtocol(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        examples = textwrap.dedent('''
            mdt-math-protocol protocol.prtcl 'G *= 1e-3' -o new_protocol.prtcl
            mdt-math-protocol p.prtcl 'G *= 1e-3; TR /= 1000; TE /= 1000'
            mdt-math-protocol p.prtcl 'del(G)'
            mdt-math-protocol p.prtcl 'TE  = 50e-3'
            mdt-math-protocol p.prtcl -a Delta.txt 'Delta = files[0]'
           ''')
        epilog = self._format_examples(doc_parser, examples)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('input_protocol', metavar='input_protocol', type=str,
                            help="The input protocol")

        parser.add_argument('expr', metavar='expr', type=str,
                            help="The expression to evaluate.")

        parser.add_argument('-o', '--output_file',
                            help='the output protocol, defaults to the input protocol.').completer = FilesCompleter()

        parser.add_argument('-a', '--additional-file', type=str, action='append', dest='additional_files',
                            help='additional file to load to be used for columns, placed in \'files\' list by index')

        return parser

    def run(self, args, extra_args):
        if args.output_file is not None:
            output_file = os.path.realpath(args.output_file)
        else:
            output_file = os.path.realpath(args.input_protocol)

        additional_files = []
        if args.additional_files:
            for file in args.additional_files:
                additional_files.append(np.genfromtxt(file))

        protocol = mdt.load_protocol(os.path.realpath(args.input_protocol))
        context_dict = {name: protocol.get_column(name) for name in protocol.column_names}

        exec(args.expr, {'np': np, 'files': additional_files}, context_dict)

        for key in context_dict:
            if is_scalar(context_dict[key]):
                context_dict[key] = np.ones(protocol.length) * context_dict[key]

        protocol = Protocol(context_dict)
        mdt.write_protocol(protocol, output_file)


def get_doc_arg_parser():
    return MathProtocol().get_documentation_arg_parser()


if __name__ == '__main__':
    MathProtocol().start()



from PyQt5.QtCore import pyqtSlot, Qt
from PyQt5.QtGui import QFontMetrics
from PyQt5.QtWidgets import QWidget
import numpy as np
from mdt.gui.maps_visualizer.actions import SetMapTitle, SetMapColormap, SetMapScale, SetMapClipping, \
    SetMapColorbarLabel
from mdt.gui.maps_visualizer.base import DataConfigModel
from mdt.gui.maps_visualizer.design.ui_MapSpecificOptions import Ui_MapSpecificOptions
from mdt.gui.maps_visualizer.design.ui_TabMapSpecific import Ui_TabMapSpecific
from mdt.gui.utils import blocked_signals, TimedUpdate, split_long_path_elements
from mdt.visualization.maps.base import DataInfo, SingleMapConfig, MapPlotConfig

__author__ = 'Robbert Harms'
__date__ = "2016-09-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class TabMapSpecific(QWidget, Ui_TabMapSpecific):

    def __init__(self, controller, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self.map_specific_tab = MapSpecificOptions(controller, self)
        self.mapSpecificOptionsPosition.addWidget(self.map_specific_tab)

        self._controller = controller
        self._controller.model_updated.connect(self.model_updated)

        self.selectedMap.currentIndexChanged.connect(
            lambda ind: self._update_map_specifics(self.selectedMap.itemData(ind, Qt.UserRole)))

    @pyqtSlot(DataConfigModel)
    def model_updated(self, model):
        config = model.get_config()
        map_names = config.maps_to_show

        with blocked_signals(self.selectedMap):
            current_selected = self.selectedMap.currentData(Qt.UserRole)

            self.selectedMap.clear()
            self.selectedMap.addItems(map_names)

            for index, map_name in enumerate(map_names):
                self.selectedMap.setItemData(index, map_name, Qt.UserRole)

                if map_name in config.map_plot_options and config.map_plot_options[map_name].title:
                    title = config.map_plot_options[map_name].title
                    self.selectedMap.setItemData(index, map_name + ' (' + title + ')', Qt.DisplayRole)

            for ind in range(self.selectedMap.count()):
                if self.selectedMap.itemData(ind, Qt.UserRole) == current_selected:
                    self.selectedMap.setCurrentIndex(ind)
                    break

        if self.selectedMap.count():
            self._update_map_specifics(self.selectedMap.currentData(Qt.UserRole))
        else:
            self._update_map_specifics(None)

    def _update_map_specifics(self, map_name):
        """Set the map specific options to reflect the settings of the given map"""
        if map_name is None:
            self.map_specific_tab.reset()
        else:
            self.map_specific_tab.use(map_name)


class MapSpecificOptions(QWidget, Ui_MapSpecificOptions):

    def __init__(self, controller, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self._controller = controller

        current_model = self._controller.get_model()

        self._current_map = None
        self.colormap.addItems(['-- Use global --'] + current_model.get_config().get_available_colormaps())
        self.colormap.currentIndexChanged.connect(self._update_colormap)
        self.data_clipping_min.valueChanged.connect(self._update_clipping_min)
        self.data_clipping_max.valueChanged.connect(self._update_clipping_max)
        self.data_scale_min.valueChanged.connect(self._update_scale_min)
        self.data_scale_max.valueChanged.connect(self._update_scale_max)

        self.data_set_use_scale.stateChanged.connect(self._set_use_scale)
        self.use_data_scale_min.stateChanged.connect(self._set_use_data_scale_min)
        self.use_data_scale_max.stateChanged.connect(self._set_use_data_scale_max)

        self.data_set_use_clipping.stateChanged.connect(self._set_use_clipping)
        self.use_data_clipping_min.stateChanged.connect(self._set_use_data_clipping_min)
        self.use_data_clipping_max.stateChanged.connect(self._set_use_data_clipping_max)

        self._title_timer = TimedUpdate(self._update_map_title)
        self.map_title.textChanged.connect(lambda: self._title_timer.add_delayed_callback(500))
        self.map_title.setFixedHeight(QFontMetrics(self.map_title.font()).lineSpacing() * 3)

        self._colorbar_label_timer = TimedUpdate(self._update_colorbar_label)
        self.data_colorbar_label.textChanged.connect(lambda : self._colorbar_label_timer.add_delayed_callback(500))
        self.data_colorbar_label.setFixedHeight(QFontMetrics(self.data_colorbar_label.font()).lineSpacing() * 3)

        self.info_Clipping.set_collapse(True)

        self._auto_enable_scale_min = False
        self._auto_enable_scale_max = False
        self._auto_enable_clipping_min = False
        self._auto_enable_clipping_max = False

        self.reset()
        self._update_scaling_delays()

    def _update_scaling_delays(self):
        if self.use_data_scale_max.isChecked():
            self.data_scale_max.set_update_delay(500)
        else:
            self.data_scale_max.set_update_delay(0)

        if self.use_data_scale_min.isChecked():
            self.data_scale_min.set_update_delay(500)
        else:
            self.data_scale_min.set_update_delay(0)

        if self.use_data_clipping_max.isChecked():
            self.data_clipping_max.set_update_delay(500)
        else:
            self.data_clipping_max.set_update_delay(0)

        if self.use_data_clipping_min.isChecked():
            self.data_clipping_min.set_update_delay(500)
        else:
            self.data_clipping_min.set_update_delay(0)

    def reset(self):
        """Set all the values to their defaults"""
        self._current_map = None
        self.colormap.setCurrentText('hot')

        with blocked_signals(self.map_title):
            self.map_title.document().setPlainText('')

        with blocked_signals(self.data_colorbar_label):
            self.data_colorbar_label.document().setPlainText('')

        with blocked_signals(self.data_clipping_min):
            self.data_clipping_min.setValue(0)

        with blocked_signals(self.data_clipping_max):
            self.data_clipping_max.setValue(0)

        with blocked_signals(self.data_scale_min):
            self.data_scale_min.setValue(0)

        with blocked_signals(self.data_scale_max):
            self.data_scale_max.setValue(0)

        with blocked_signals(self.use_data_clipping_min):
            self.use_data_clipping_min.setChecked(False)

        with blocked_signals(self.use_data_clipping_max):
            self.use_data_clipping_max.setChecked(False)

        with blocked_signals(self.use_data_scale_min):
            self.use_data_scale_min.setChecked(False)

        with blocked_signals(self.use_data_scale_max):
            self.use_data_scale_max.setChecked(False)

        self.info_file_location.setText('-')
        self.info_maximum.setText('-')
        self.info_minimum.setText('-')
        self.info_shape.setText('-')

    def _get_mask(self, data_info, map_name):
        current_model = self._controller.get_model()
        plot_config = current_model.get_config()

        def get_map_attr():
            if map_name in plot_config.map_plot_options:
                value = getattr(plot_config.map_plot_options[map_name], 'mask_name')
                if value is not None:
                    return value
            return plot_config.mask_name

        mask_name = get_map_attr()

        if mask_name is not None:
            return data_info.get_map_data(mask_name)
        return None

    def use(self, map_name):
        """Load the settings of the given map"""
        self._current_map = map_name

        current_model = self._controller.get_model()

        try:
            map_info = current_model.get_config().map_plot_options[map_name]
        except KeyError:
            map_info = SingleMapConfig()

        data_info = current_model.get_data()
        vmin, vmax = data_info.get_single_map_info(map_name).min_max(mask=self._get_mask(data_info, map_name))

        with blocked_signals(self.map_title):
            self.map_title.document().setPlainText(map_info.title if map_info.title else '')

        with blocked_signals(self.data_colorbar_label):
            self.data_colorbar_label.document().setPlainText(map_info.colorbar_label if map_info.colorbar_label else '')

        with blocked_signals(self.colormap):
            if map_info.colormap is None:
                self.colormap.setCurrentIndex(0)
            else:
                self.colormap.setCurrentText(map_info.colormap)

        with blocked_signals(self.data_clipping_min):
            self.data_clipping_min.setValue(map_info.clipping.vmin)

        with blocked_signals(self.data_clipping_max):
            self.data_clipping_max.setValue(map_info.clipping.vmax)

        with blocked_signals(self.data_scale_min):
            self.data_scale_min.setValue(map_info.scale.vmin)

        with blocked_signals(self.data_scale_max):
            self.data_scale_max.setValue(map_info.scale.vmax)

        with blocked_signals(self.data_set_use_scale):
            self.data_set_use_scale.setChecked(map_info.scale.use_min or map_info.scale.use_max)

        with blocked_signals(self.data_set_use_clipping):
            self.data_set_use_clipping.setChecked(map_info.clipping.use_min or map_info.clipping.use_max)

        with blocked_signals(self.use_data_clipping_min):
            self.use_data_clipping_min.setChecked(map_info.clipping.use_min)

        with blocked_signals(self.use_data_clipping_max):
            self.use_data_clipping_max.setChecked(map_info.clipping.use_max)

        with blocked_signals(self.use_data_scale_min):
            self.use_data_scale_min.setChecked(map_info.scale.use_min)

        with blocked_signals(self.use_data_scale_max):
            self.use_data_scale_max.setChecked(map_info.scale.use_max)

        map_filename = data_info.get_file_path(map_name)
        if map_filename:
            self.info_file_location.setText(split_long_path_elements(map_filename, 25))

        self.info_minimum.setText(str(vmin))
        self.info_maximum.setText(str(vmax))
        self.info_has_nan.setText('Yes' if data_info.get_single_map_info(map_name).has_nan() else 'No')

        self.info_shape.setText(str(data_info.get_single_map_info(map_name).shape))

    def _get_current_map_config(self):
        current_model = self._controller.get_model()
        current_config = current_model.get_config()
        current_map_config = current_config.map_plot_options.get(self._current_map, SingleMapConfig())
        return current_map_config

    @pyqtSlot()
    def _update_map_title(self):
        string = self.map_title.toPlainText()
        if self._current_map:
            if string == '':
                string = None
            self._controller.apply_action(SetMapTitle(self._current_map, string))

    @pyqtSlot()
    def _update_colorbar_label(self):
        string = self.data_colorbar_label.toPlainText()
        if self._current_map:
            if string == '':
                string = None
            self._controller.apply_action(SetMapColorbarLabel(self._current_map, string))

    @pyqtSlot(int)
    def _update_colormap(self, index):
        if self._current_map:
            if index == 0:
                self._controller.apply_action(SetMapColormap(self._current_map, None))
            else:
                self._controller.apply_action(SetMapColormap(self._current_map, self.colormap.itemText(index)))

    @pyqtSlot(float)
    def _update_scale_min(self, value):
        if self._current_map:
            current_scale = self._get_current_map_config().scale
            kwargs = dict(vmin=value)

            if current_scale.use_max and current_scale.use_min and value > current_scale.vmax:
                kwargs.update(dict(use_min=False))
                self._auto_enable_scale_min = True
            elif self._auto_enable_scale_min and value < current_scale.vmax:
                kwargs.update(dict(use_min=True))
                self._auto_enable_scale_min = False

            new_scale = current_scale.get_updated(**kwargs)
            self._controller.apply_action(SetMapScale(self._current_map, new_scale))

    @pyqtSlot(float)
    def _update_scale_max(self, value):
        if self._current_map:
            current_scale = self._get_current_map_config().scale
            kwargs = dict(vmax=value)

            if current_scale.use_max and current_scale.use_min and value < current_scale.vmin:
                kwargs.update(dict(use_max=False))
                self._auto_enable_scale_max = True
            elif self._auto_enable_scale_max and value > current_scale.vmin:
                kwargs.update(dict(use_max=True))
                self._auto_enable_scale_max = False

            new_scale = current_scale.get_updated(**kwargs)
            self._controller.apply_action(SetMapScale(self._current_map, new_scale))

    @pyqtSlot(int)
    def _set_use_scale(self, check_state):
        use_scale = check_state == 2
        if self._current_map:
            current_scale = self._get_current_map_config().scale
            if use_scale and current_scale.vmax < current_scale.vmin:
                new_scale = current_scale.get_updated(use_min=use_scale, use_max=use_scale, vmax=current_scale.vmin)
            else:
                new_scale = current_scale.get_updated(use_min=use_scale, use_max=use_scale)
            self._controller.apply_action(SetMapScale(self._current_map, new_scale))
        self._update_scaling_delays()

    @pyqtSlot(int)
    def _set_use_data_scale_min(self, check_state):
        use_scale = check_state == 2
        if self._current_map:
            if use_scale and self._get_current_map_config().scale.use_max:
                self._set_use_scale(2)
            else:
                new_scale = self._get_current_map_config().scale.get_updated(use_min=use_scale)
                self._controller.apply_action(SetMapScale(self._current_map, new_scale))

    @pyqtSlot(int)
    def _set_use_data_scale_max(self, check_state):
        use_scale = check_state == 2
        if self._current_map:
            if use_scale and self._get_current_map_config().scale.use_min:
                self._set_use_scale(2)
            else:
                new_scale = self._get_current_map_config().scale.get_updated(use_max=use_scale)
                self._controller.apply_action(SetMapScale(self._current_map, new_scale))

    @pyqtSlot(float)
    def _update_clipping_min(self, value):
        if self._current_map:
            current_clipping = self._get_current_map_config().clipping
            kwargs = dict(vmin=value)

            if current_clipping.use_max and current_clipping.use_min and value > current_clipping.vmax:
                kwargs.update(dict(use_min=False))
                self._auto_enable_scale_min = True
            elif self._auto_enable_scale_min and value < current_clipping.vmax:
                kwargs.update(dict(use_min=True))
                self._auto_enable_scale_min = False

            new_clipping = current_clipping.get_updated(**kwargs)
            self._controller.apply_action(SetMapClipping(self._current_map, new_clipping))

    @pyqtSlot(float)
    def _update_clipping_max(self, value):
        if self._current_map:
            current_clipping = self._get_current_map_config().clipping
            kwargs = dict(vmax=value)

            if current_clipping.use_max and current_clipping.use_min and value < current_clipping.vmin:
                kwargs.update(dict(use_max=False))
                self._auto_enable_scale_max = True
            elif self._auto_enable_scale_max and value > current_clipping.vmin:
                kwargs.update(dict(use_max=True))
                self._auto_enable_scale_max = False

            new_clipping = current_clipping.get_updated(**kwargs)
            self._controller.apply_action(SetMapClipping(self._current_map, new_clipping))

    @pyqtSlot(int)
    def _set_use_clipping(self, check_state):
        use_clipping = check_state == 2
        if self._current_map:
            current_clipping = self._get_current_map_config().clipping
            if use_clipping and current_clipping.vmax < current_clipping.vmin:
                new_clipping = current_clipping.get_updated(use_min=use_clipping, use_max=use_clipping,
                                                            vmax=current_clipping.vmin)
            else:
                new_clipping = current_clipping.get_updated(use_min=use_clipping, use_max=use_clipping)
            self._controller.apply_action(SetMapClipping(self._current_map, new_clipping))
        self._update_scaling_delays()

    @pyqtSlot(int)
    def _set_use_data_clipping_min(self, check_state):
        use_clipping = check_state == 2
        if self._current_map:
            if use_clipping and self._get_current_map_config().clipping.use_max:
                self._set_use_clipping(2)
            else:
                new_clipping = self._get_current_map_config().clipping.get_updated(use_min=use_clipping)
                self._controller.apply_action(SetMapClipping(self._current_map, new_clipping))

    @pyqtSlot(int)
    def _set_use_data_clipping_max(self, check_state):
        use_clipping = check_state == 2
        if self._current_map:
            if use_clipping and self._get_current_map_config().clipping.use_min:
                self._set_use_clipping(2)
            else:
                new_clipping = self._get_current_map_config().clipping.get_updated(use_max=use_clipping)
                self._controller.apply_action(SetMapClipping(self._current_map, new_clipping))



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Estimate the noise standard deviation of the Gaussian noise in the original complex image domain.

The result is a single floating point number with the noise std. for every voxel. It uses the estimation routines
from the components folders for the estimation. The estimation is the same as the one used in mdt-model-fit, but
since the noise std estimation depends on the mask used, it is better to call this function beforehand with a
complete brain mask. Later, the mdt-model-fit routine can be called on smaller masks with as noise std the value
from this function.
"""
import argparse
import os
import mdt
from argcomplete.completers import FilesCompleter
from mdt.lib.shell_utils import BasicShellApplication
from mot.lib import cl_environments
import textwrap


__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NoiseStdEstimation(BasicShellApplication):

    def __init__(self):
        super().__init__()
        self.available_devices = list((ind for ind, env in
                                       enumerate(cl_environments.CLEnvironmentFactory.smart_device_selection())))

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        examples = textwrap.dedent('''
            mdt-estimate-noise-std data.nii.gz data.prtcl full_mask.nii.gz
        ''')
        epilog = self._format_examples(doc_parser, examples)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('dwi',
                            action=mdt.lib.shell_utils.get_argparse_extension_checker(['.nii', '.nii.gz', '.hdr', '.img']),
                            help='the diffusion weighted image').completer = FilesCompleter(['nii', 'gz', 'hdr', 'img'],
                                                                                            directories=False)
        parser.add_argument(
            'protocol', action=mdt.lib.shell_utils.get_argparse_extension_checker(['.prtcl']),
            help='the protocol file, see mdt-create-protocol').completer = FilesCompleter(['prtcl'],
                                                                                            directories=False)
        parser.add_argument('mask',
                            action=mdt.lib.shell_utils.get_argparse_extension_checker(['.nii', '.nii.gz', '.hdr', '.img']),
                            help='the (brain) mask to use').completer = FilesCompleter(['nii', 'gz', 'hdr', 'img'],
                                                                               directories=False)

        parser.add_argument('--estimator', '-e', type=str, help='The name of the estimation routine to use')

        return parser

    def run(self, args, extra_args):
        input_data = mdt.load_input_data(os.path.realpath(args.dwi),
                                         os.path.realpath(args.protocol),
                                         os.path.realpath(args.mask))

        estimator = None
        if args.estimator:
            estimator = mdt.load_component('noise_std_estimators', args.estimator)

        with mdt.with_logging_to_debug():
            noise_std = mdt.estimate_noise_std(input_data, estimator=estimator)
            print(noise_std)


def get_doc_arg_parser():
    return NoiseStdEstimation().get_documentation_arg_parser()


if __name__ == '__main__':
    NoiseStdEstimation().start()



import os
from textwrap import dedent

from mdt.lib.nifti import load_nifti
from PyQt5.QtCore import pyqtSlot, QObject, pyqtSignal
from PyQt5.QtWidgets import QFileDialog

from mdt.visualization.maps.base import SimpleDataInfo, MapPlotConfig
from mdt.gui.maps_visualizer.main import start_gui
from mdt.gui.model_fit.design.ui_generate_roi_mask_tab import Ui_GenerateROIMaskTabContent
from mdt.gui.utils import function_message_decorator, image_files_filters, MainTab, get_script_file_header_text
from mdt.utils import split_image_path, write_slice_roi

__author__ = 'Robbert Harms'
__date__ = "2016-06-26"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GenerateROIMaskTab(MainTab, Ui_GenerateROIMaskTabContent, QObject):

    def __init__(self, shared_state, computations_thread):
        super().__init__()
        self._shared_state = shared_state
        self._computations_thread = computations_thread
        self._generate_mask_worker = GenerateROIMaskWorker()

    def setupUi(self, tab_content):
        super().setupUi(tab_content)

        self.selectMaskButton.clicked.connect(lambda: self._select_mask())
        self.selectOutputFileInput.clicked.connect(lambda: self._select_output_file())

        self.viewButton.clicked.connect(self.view_mask)
        self.generateButton.clicked.connect(self.generate_roi_mask)

        self.selectedMaskText.textChanged.connect(self._check_enable_action_buttons)
        self.selectedOutputFileText.textChanged.connect(self._check_enable_action_buttons)
        self.selectedMaskText.textChanged.connect(self.mask_file_changed)

        self.dimensionInput.valueChanged.connect(self.update_dimension)
        self.sliceInput.valueChanged.connect(self.update_slice_index)

    def _select_mask(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedMaskText.text() != '':
            initial_dir = self.selectedMaskText.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the brain mask', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if os.path.isfile(open_file):
            self.selectedMaskText.setText(open_file)
            self.mask_file_changed()
            self._shared_state.base_dir = os.path.dirname(open_file)

    def _select_output_file(self):
        output_file_name, used_filter = QFileDialog().getSaveFileName(
            caption='Select the output file', directory=self._shared_state.base_dir,
            filter=';;'.join(image_files_filters))

        if output_file_name:
            self.selectedOutputFileText.setText(output_file_name)

    def _check_enable_action_buttons(self):
        self.generateButton.setEnabled(os.path.isfile(self.selectedMaskText.text()) and
                                       os.path.isdir(os.path.dirname(self.selectedOutputFileText.text())))
        self.viewButton.setEnabled(os.path.isfile(self.selectedMaskText.text()) and
                                   os.path.isfile(self.selectedOutputFileText.text()))

    def view_mask(self):
        data = SimpleDataInfo({'Original mask': load_nifti(self.selectedMaskText.text()).get_data(),
                               'Slice mask': load_nifti(self.selectedOutputFileText.text()).get_data()})

        config = MapPlotConfig()
        config.dimension = self.dimensionInput.value()
        config.slice_index = self.sliceInput.value()
        config.maps_to_show = ['Original mask', 'Slice mask']

        start_gui(data=data, config=config, app_exec=False)

    def generate_roi_mask(self):
        kwargs = dict(mask=self.selectedMaskText.text(), output=self.selectedOutputFileText.text(),
                      dimension=self.dimensionInput.value(), slice=self.sliceInput.value())
        self._generate_mask_worker.set_args(**kwargs)
        self._computations_thread.start()
        self._generate_mask_worker.moveToThread(self._computations_thread)

        self._generate_mask_worker.starting.connect(self._computations_thread.starting)
        self._generate_mask_worker.finished.connect(self._computations_thread.finished)

        self._generate_mask_worker.starting.connect(lambda: self.generateButton.setEnabled(False))
        self._generate_mask_worker.finished.connect(lambda: self.generateButton.setEnabled(True))
        self._generate_mask_worker.finished.connect(lambda: self.viewButton.setEnabled(True))

        image_path = split_image_path(self.selectedOutputFileText.text())
        script_basename = os.path.join(image_path[0], 'scripts', 'create_roi_mask_' + image_path[1])
        if not os.path.isdir(os.path.join(image_path[0], 'scripts')):
            os.makedirs(os.path.join(image_path[0], 'scripts'))

        self._generate_mask_worker.finished.connect(
            lambda: self._write_python_script_file(script_basename + '.py', **kwargs))

        self._generate_mask_worker.finished.connect(
            lambda: self._write_bash_script_file(script_basename + '.sh', **kwargs))

        self._generate_mask_worker.starting.emit()

    @pyqtSlot(int)
    def update_dimension(self, value):
        if os.path.isfile(self.selectedMaskText.text()):
            self.update_slice_selector()
            self.update_output_file_text()

    @pyqtSlot(int)
    def update_slice_index(self, value):
        self.update_output_file_text()

    def mask_file_changed(self):
        self.dimensionInput.setValue(2)
        self.update_slice_selector()
        self.update_output_file_text()

    def update_slice_selector(self):
        if os.path.isfile(self.selectedMaskText.text()):
            dimension_max = load_nifti(self.selectedMaskText.text()).shape[self.dimensionInput.value()]
            self.sliceInput.setMaximum(dimension_max)
            self.sliceInput.setValue(dimension_max // 2.0)
            self.maxSliceLabel.setText(str(dimension_max))
        else:
            self.sliceInput.setValue(0)
            self.maxSliceLabel.setText('x')

    def update_output_file_text(self):
        if os.path.isfile(self.selectedMaskText.text()):
            folder, basename, ext = split_image_path(self.selectedMaskText.text())
            folder_base = os.path.join(folder, basename)

            if self.selectedOutputFileText.text() == '':
                self.selectedOutputFileText.setText('{}_{}_{}.nii.gz'.format(folder_base,
                                                                             self.dimensionInput.value(),
                                                                             self.sliceInput.value()))
            elif self.selectedOutputFileText.text()[0:len(folder_base)] == folder_base:
                self.selectedOutputFileText.setText('{}_{}_{}.nii.gz'.format(folder_base,
                                                                             self.dimensionInput.value(),
                                                                             self.sliceInput.value()))

    def _write_python_script_file(self, output_file, **kwargs):
        with open(output_file, 'w') as f:
            f.write('#!/usr/bin/env python')
            f.write(dedent('''

                {header}

                import mdt

                mdt.write_slice_roi(
                    {mask!r},
                    {dimension!r},
                    {slice!r},
                    {output!r},
                    overwrite_if_exists=True)

            ''').format(header=get_script_file_header_text({'Purpose': 'Generated a slice ROI mask'}),
                        **kwargs))

    def _write_bash_script_file(self, output_file, **kwargs):
        with open(output_file, 'w') as f:
            f.write('#!/usr/bin/env bash')
            f.write(dedent('''

                {header}

                mdt-create-roi-slice "{mask}" -d {dimension} -s {slice} -o "{output}"
            ''').format(header=get_script_file_header_text({'Purpose': 'Generated a slice ROI mask'}),
                        **kwargs))


class GenerateROIMaskWorker(QObject):

    starting = pyqtSignal()
    finished = pyqtSignal()

    def __init__(self):
        super().__init__()
        self.starting.connect(self.run)
        self._args = []
        self._kwargs = {}

    def set_args(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @function_message_decorator('Started with generating a slice ROI', 'Finished generating a slice ROI')
    def run(self):
        write_slice_roi(self._kwargs['mask'], self._kwargs['dimension'],
                        self._kwargs['slice'], self._kwargs['output'], overwrite_if_exists=True)
        self.finished.emit()



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""This script prints information about the available devices on your computer."""
import argparse
import textwrap
from mdt.lib.shell_utils import BasicShellApplication
from mot.lib import cl_environments

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ListDevices(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('-l', '--long', action='store_true', help='print all info about the devices')
        return parser

    def run(self, args, extra_args):
        for ind, env in enumerate(cl_environments.CLEnvironmentFactory.smart_device_selection()):
            print('Device {}:'.format(ind))
            if args.long:
                print(repr(env))
            else:
                print(str(env))


def get_doc_arg_parser():
    return ListDevices().get_documentation_arg_parser()


if __name__ == '__main__':
    ListDevices().start()



from mdt import LibraryFunctionTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-08-27'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class BinghamNormalization_3x3(LibraryFunctionTemplate):
    description = '''
        Computes 1F1(1/2; 3/2; v), the confluent hypergeometric function of the first kind for a 3x3 matrix [1].

        This computes the normalization factor for the Bingham distribution of a 3x3 matrix by means of a saddlepoint
        approximation [2].

        The argument v is a vector of eigenvalues of a 3x3 matrix V for which you want to compute the Bingham 
        normalization factor.

        Args:
            v: the eigenvalues of the matrix

        References:
        [1] Mardia, K.V., Jupp, P.E., 2000. Distributions on spheres. Directional Statistics. 
            John Wiley & Sons, pp. 159–192  
        [2] Kume A, Wood ATA. Saddlepoint approximations for the Bingham and Fisher-Bingham normalising constants. 
            Biometrika. 2005;92(2):465-476. doi:10.1093/biomet/92.2.465.
    '''
    return_type = 'double'
    parameters = [('mot_float_type*', 'v')]
    dependencies = ['solve_cubic_pol_real']
    cl_code = '''
        if(v[0] == 0 && v[1] == 0 && v[2] == 0){
            return 1;
        }

        double coef_roots[4] = {
            2, 
            -2*v[0] - 2*v[1] - 2*v[2] + 3,
            2*v[0]*v[1] + 2*v[0]*v[2] - 2*v[0] + 2*v[1]*v[2] - 2*v[1] - 2*v[2], 
            -2*v[0]*v[1]*v[2] + v[0]*v[1] + v[0]*v[2] + v[1]*v[2]
        };
        int nmr_roots = solve_cubic_pol_real(coef_roots, coef_roots);
        double t = min(coef_roots[0], min(coef_roots[1], coef_roots[2]));

        double prod = 1;
        for(int i = 0; i < 3; i++){
            if(v[i] - t > 0){
                prod *= 1/sqrt(v[i] - t);
            }
        }

        double K2 = 0.5 * (1.0 / pown(v[0]-t, 2) + 1.0 / pown(v[1]-t, 2) + 1.0 / pown(v[2]-t, 2));
        double K3 =       (1.0 / pown(v[0]-t, 3) + 1.0 / pown(v[1]-t, 3) + 1.0 / pown(v[2]-t, 3));
        double K4 = 3   * (1.0 / pown(v[0]-t, 4) + 1.0 / pown(v[1]-t, 4) + 1.0 / pown(v[2]-t, 4));

        double T = (1/8.) * (K4 / pown(K2, 2)) - (5/24.) * (pown(K3, 2)/pown(K2, 3));  
        return M_PI * sqrt(2.0/K2) * prod * exp(-t + T);        
    '''



from PyQt5.QtCore import QObject, pyqtSignal
from mdt.visualization.maps.base import MapPlotConfig, SimpleDataInfo


class DataConfigModel(object):
    """The model contains all the state information for viewing the maps, both the data and the configuration."""

    def get_data(self):
        """Get the current data.

        Returns:
            mdt.visualization.maps.base.DataInfo: the information about the data we are viewing
        """
        raise NotImplementedError()

    def get_config(self):
        """Get the current configuration.

        Returns:
            mdt.visualization.maps.base.MapPlotConfig: the visualization configuration.
        """
        raise NotImplementedError()


class SimpleDataConfigModel(DataConfigModel):

    def __init__(self, data, config):
        """The model contains all the state information of the current viewing.

        Args:
            data (mdt.visualization.maps.base.DataInfo): the data information object
            config (mdt.visualization.maps.base.MapPlotConfig): the configuration
        """
        self._data = data
        self._config = config

    def get_data(self):
        return self._data

    def get_config(self):
        return self._config


class Controller(object):

    def __init__(self):
        """Controller interface"""
        super().__init__()

    def get_model(self):
        """Get the model the view should represent.

        Returns:
            DataConfigModel: the model the view should represent. This is also the model the actions can use for updating
                the model.
        """
        raise NotImplementedError()

    def set_data(self, data_info, config=None):
        """Set new data to visualize.

        Args:
            data_info (mdt.visualization.maps.base.DataInfo): the new data to visualize
            config (mdt.visualization.maps.base.MapPlotConfig): the new configuration for the data
                If given, we will display the new data immediately with the given config
        """

    def apply_action(self, action, store_in_history=True):
        """Apply a new configuration action.

        If there is no difference between the current config and the one generated by this new action, the
        action will not be stored in history and will not need to be applied.

        Args:
            action (mdt.gui.maps_visualizer.base.ConfigAction): the configuration action to add and apply
            store_in_history (boolean): if this action should be stored in the history or not
        """

    def undo(self):
        """Undo a previous configuration action"""

    def redo(self):
        """Reapply a previously undone configuration action"""

    def has_undo(self):
        """Check if this controller has an undo action available.

        Returns:
            boolean: True if an undo action is available.
        """

    def has_redo(self):
        """Check if this controller has an redo action available.

        Returns:
            boolean: True if an redo action is available.
        """


class QtController(Controller, QObject):

    model_updated = pyqtSignal(DataConfigModel)

    def __init__(self):
        super().__init__()
        self._current_model = SimpleDataConfigModel(SimpleDataInfo({}), MapPlotConfig())
        self._actions_history = []
        self._redoable_actions = []

    def set_data(self, data_info, config=None):
        config = config or MapPlotConfig()
        if not isinstance(config, MapPlotConfig):
            config = MapPlotConfig.from_dict(config.to_dict())

        self._current_model = SimpleDataConfigModel(data_info, config)

        if data_info.get_map_names():
            max_dim = data_info.get_max_dimension()
            if max_dim < config.dimension:
                config.dimension = max_dim

        config.maps_to_show = list(filter(lambda k: k in data_info.get_map_names(), config.maps_to_show))

        self._actions_history = []
        self._redoable_actions = []
        self.model_updated.emit(self._current_model)

    def get_model(self):
        return self._current_model

    def apply_action(self, action, store_in_history=True):
        new_model = action.apply(self._current_model)
        is_applied = self._apply_new_model(new_model)
        if is_applied:
            if store_in_history:
                self._actions_history.append(action)
                self._redoable_actions = []
            self.model_updated.emit(self._current_model)

    def undo(self):
        if len(self._actions_history):
            action = self._actions_history.pop()
            self._apply_new_model(action.unapply())
            self._redoable_actions.append(action)
            self.model_updated.emit(self._current_model)

    def redo(self):
        if len(self._redoable_actions):
            action = self._redoable_actions.pop()
            self._apply_new_model(action.apply(self._current_model))
            self._actions_history.append(action)
            self.model_updated.emit(self._current_model)

    def has_undo(self):
        return len(self._actions_history) > 0

    def has_redo(self):
        return len(self._redoable_actions) > 0

    def _apply_new_model(self, new_model):
        """Apply the current configuration.

        Args:
            new_model (DataConfigModel): the new model to apply

        Returns:
            bool: if the model was applied or not. If the difference with the current configuration
                and the old one is None, we return False. Else we return True.
        """
        new_config = new_model.get_config()
        if new_config.validate(new_model.get_data())\
                or self._current_model.get_data() != new_model.get_data():
            self._current_model = new_model
            return True
        return False



import ast
import inspect
from textwrap import indent, dedent

__author__ = 'Robbert Harms'
__date__ = '2017-07-24'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class TemplateModifier(object):

    def __init__(self, template):
        """Given a template instance, this class can update the properties of the template and write that to a file.

        This will overwrite the template definition in the source file where the template was originally defined. If you
        desire a different location then first create a new file for the template.

        Args:
            template (ComponentTemplate): the template to update
        """
        self._source = inspect.getsource(template)
        self._filename = inspect.getfile(template)

        source_lines = inspect.getsourcelines(template)
        self._template_start = source_lines[1] - 1
        self._template_end = len(source_lines[0]) + self._template_start

    def update(self, property_name, source_code_str):
        """Update the given property with the given source code.

        This does not write the results to file immediately, rather, this updates an internal buffer with the
        updated source code. To write to file use :meth:`write_to_file`.

        Args:
            property_name (str): the property (attribute or function) to update
            source_code_str (str): the updated source code for the property
        """
        try:
            start, end = self._fine_property_definition(property_name)
            source_lines = self._source.split('\n')
            new_source = '\n'.join(source_lines[:start])
            new_source += '\n' + indent(dedent(source_code_str.strip()), '\t') + '\n'
            new_source += '\n'.join(source_lines[end:])
        except ValueError:
            new_source = self._source + '\n' + indent(dedent(source_code_str.strip()), '\t') + '\n'
        self._source = new_source.rstrip('\n').replace('\t', '    ')

    def get_source(self):
        """Return the current source code buffer.

        Returns:
            str: the updated source code
        """
        return self._source

    def write_to_file(self):
        """Write the updated source code to file.

        This can be called repetitively after calling :meth:`update` since this method will keep track of the current
        position in the file.
        """
        with open(self._filename, 'r') as f:
            lines = [line.rstrip('\n') for line in f]

        start_lines = lines[:self._template_start]
        new_source_lines = self._source.split('\n')
        end_lines = lines[self._template_end:]

        lines = start_lines + new_source_lines + end_lines

        with open(self._filename, 'w') as f:
            f.writelines('{}\n'.format(l) for l in lines)

        self._template_start = len(start_lines)
        self._template_end = len(start_lines) + len(new_source_lines)

    def _fine_property_definition(self, property_name):
        """Find the lines in the source code that contain this property's name and definition.

        This function can find both attribute assignments as well as methods/functions.

        Args:
            property_name (str): the name of the property to look up in the template definition

        Returns:
            tuple: line numbers for the start and end of the attribute definition
        """
        for node in ast.walk(ast.parse(self._source)):
            if isinstance(node, ast.Assign) and node.targets[0].id == property_name:
                return node.targets[0].lineno - 1, self._get_node_line_end(node)
            elif isinstance(node, ast.FunctionDef) and node.name == property_name:
                return node.lineno - 1, self._get_node_line_end(node)
        raise ValueError('The requested node could not be found.')

    def _get_node_line_end(self, node):
        """Get the last line of the given node.

        This function can recurse if the given node is a complex node (like a FunctionDef node).

        Args:
            node (ast.Node): a node of the AST

        Returns:
            int: the last line of the statements in the given node
        """
        if isinstance(node, ast.Assign):
            return node.value.lineno
        elif isinstance(node, ast.FunctionDef):
            return self._get_node_line_end(node.body[-1])



"""Definitions of the free parameters.

The free parameters are meant to be used for parameters that one wants to optimize. They can be fixed to a certain
value to disable them from being optimized in a given situation, but they remain classified as 'optimizable' parameters.
"""

import numpy as np
from mdt import FreeParameterTemplate
from mdt.model_building.parameter_functions.priors import UniformWithinBoundsPrior, ARDBeta, ARDGaussian


__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class s0(FreeParameterTemplate):

    init_value = 1e4
    lower_bound = 0
    upper_bound = 1e10
    parameter_transform = 'Clamp'
    sampling_proposal_std = 10.0


class w(FreeParameterTemplate):

    init_value = 0.5
    lower_bound = 0
    upper_bound = 1
    parameter_transform = 'CosSqrClamp'
    sampling_proposal_std = 0.01
    sampling_prior = UniformWithinBoundsPrior()
    numdiff_info = {'scale_factor': 10}


class w_ard_beta(w):
    """Subclasses the weight to add a Beta prior for in use with Automatic Relevance Detection during sample."""
    sampling_prior = ARDBeta()


class w_ard_gaussian(w):
    """Subclasses the weight to add a Gaussian prior for in use with Automatic Relevance Detection during sample."""
    sampling_prior = ARDGaussian()


class T1(FreeParameterTemplate):

    init_value = 0.02
    lower_bound = 1e-5
    upper_bound = 4.0
    parameter_transform = 'Clamp'
    sampling_proposal_std = 0.0001


class T2(FreeParameterTemplate):

    init_value = 0.01
    lower_bound = 1e-5
    upper_bound = 2.0
    parameter_transform = 'Clamp'
    sampling_proposal_std = 0.0001


class T2_star(FreeParameterTemplate):

    init_value = 0.01
    lower_bound = 0.0
    upper_bound = 1.0
    parameter_transform = 'Clamp'
    sampling_proposal_std = 0.0001


class E1(FreeParameterTemplate):
    """This parameter is defined *only* for linear decay T1 fitting in GRE data *with* TR constant.

    This parameter is also defined in the SSFP equation. However, in SSFP this parameter is from the protocol (!)
        E1 = exp( -TR / T1 ).
    After estimation of this parameter, T1 can be recovered by applying the next equation:
        -TR / log( E1 ).
    """

    init_value = 0.37
    lower_bound = 0.0
    upper_bound = 1.0
    parameter_transform = 'Clamp'
    sampling_proposal_std = 0.0001


class R1(FreeParameterTemplate):
    """R1 = 1/T1, for linear T1Dec or other models. """

    init_value = 2
    lower_bound = 0.25
    upper_bound = 100.0
    parameter_transform = 'Clamp'
    sampling_proposal_std = 0.0001


class R2(FreeParameterTemplate):
    """R2 = 1/T2, for linear T2Dec or other models."""

    init_value = 5
    lower_bound = 0.5
    upper_bound = 500.0
    parameter_transform = 'Clamp'
    sampling_proposal_std = 0.0001


class R2s(FreeParameterTemplate):
    """R2s = 1/T2s, for lineaR T2sDec or other models."""

    init_value = 10
    lower_bound = 1
    upper_bound = 50.0
    parameter_transform = 'Clamp'
    sampling_proposal_std = 0.0001


class theta(FreeParameterTemplate):
    """The inclination/polar angle.

    This parameter is limited between [0, pi] but with modulus pi. That is, 0 == pi and this parameter should be
    allowed to wrap around pi.

    """
    init_value = np.pi / 2.0
    lower_bound = 0
    upper_bound = np.pi
    parameter_transform = 'AbsModPi'
    sampling_proposal_std = 0.1
    sampling_proposal_modulus = np.pi
    numdiff_info = {'use_bounds': False}


class phi(FreeParameterTemplate):
    """The azimuth angle.

    We limit this parameter between [0, pi] making us only use (together with theta between [0, pi]) only the right
    hemisphere. This is possible since diffusion is symmetric and works fine during optimization. For calculating the
    numerical derivative we can let phi rotate around 2*pi again.

    During sample the results can clip to pi since the standard formula for transforming spherical coordinates to
    cartesian coordinates defines phi to be in the range [0, 2*pi]. This is both a problem and a blessing. The problem
    is that the samples will not wrap nicely around pi, the blessing is that we prevent a bimodal distribution in phi.
    Not wrapping around pi is not much of a problem though, as the sampler can easily sample only half of a gaussian
    if the optimal parameter is around zero or pi.
    """
    init_value = np.pi / 2.0
    lower_bound = 0
    upper_bound = np.pi
    parameter_transform = 'AbsModPi'
    sampling_proposal_std = 0.1
    sampling_proposal_modulus = np.pi
    numdiff_info = {'use_bounds': False}


class psi(FreeParameterTemplate):
    """The rotation angle for use in cylindrical models.

    This parameter can be used to rotate a vector around another vector, as is for example done in the Tensor model.
    """
    init_value = np.pi / 2.0
    lower_bound = 0
    upper_bound = np.pi
    parameter_transform = 'AbsModPi'
    sampling_proposal_modulus = np.pi
    sampling_proposal_std = 0.1
    numdiff_info = {'use_bounds': False}


class d(FreeParameterTemplate):

    init_value = 1.7e-9
    lower_bound = 1e-11
    upper_bound = 1.0e-8
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class dperp0(FreeParameterTemplate):

    init_value = 1.7e-10
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class dperp1(FreeParameterTemplate):

    init_value = 1.7e-11
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class R(FreeParameterTemplate):
    init_value = 1.0e-6
    lower_bound = 1e-7
    upper_bound = 20e-6
    parameter_transform = 'CosSqrClamp'
    sampling_proposal_std = 1e-7


class kappa(FreeParameterTemplate):
    """The kappa parameter of the Watson distribution in the NODDI model.

    The NODDI model computes the spherical harmonic (SH) coefficients of the Watson distribution with the
    concentration parameter k (kappa) up to the 12th order.

    Truncating at the 12th order gives good approximation for kappa up to 64, as such we define kappa to be between
    zero and 64.
    """
    init_value = 1
    lower_bound = 0
    upper_bound = 64
    parameter_transform = 'CosSqrClamp'
    sampling_proposal_std = 0.1
    numdiff_info = {'max_step': 0.1}


# for use in ExpT1DecSTEAM model. It is assumed for ex-vivo values. For in-vivo use d instead.
class d_exvivo(FreeParameterTemplate):

    init_value = 5.0e-10
    lower_bound = 0.0
    upper_bound = 1.0e-8
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-11
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class time_dependent_characteristic_coefficient(FreeParameterTemplate):
    """The time dependent characteristic as used in the TimeDependentZeppelin model. Values are in m^2."""
    init_value = 1e-6
    lower_bound = 1e-7
    upper_bound = 1e-5
    parameter_transform = 'CosSqrClamp'
    sampling_proposal_std = 1e-7


class d_bulk(FreeParameterTemplate):

    init_value = 0.e-9
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


# the following parameters are part of the non-parametric Tensor
# (Tensor in which a upper triangular D matrix is optimized directly)
class Tensor_D_00(FreeParameterTemplate):

    init_value = 0.3e-9
    lower_bound = 0
    upper_bound = 5e-9
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class Tensor_D_11(FreeParameterTemplate):

    init_value = 0.3e-9
    lower_bound = 0
    upper_bound = 5e-9
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class Tensor_D_22(FreeParameterTemplate):

    init_value = 1.2e-9
    lower_bound = 0
    upper_bound = 5e-9
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class Tensor_D_01(FreeParameterTemplate):

    init_value = 0
    lower_bound = -1e-9
    upper_bound = 1e-9
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class Tensor_D_02(FreeParameterTemplate):

    init_value = 0
    lower_bound = -1e-9
    upper_bound = 1e-9
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}


class Tensor_D_12(FreeParameterTemplate):

    init_value = 0
    lower_bound = -1e-9
    upper_bound = 1e-9
    parameter_transform = 'SinSqrClamp'
    sampling_proposal_std = 1e-10
    numdiff_info = {'scale_factor': 1e10, 'use_upper_bound': False}



import numpy as np
from mot.sample.base import AbstractRWMSampler
from mot.lib.kernel_data import Array

__author__ = 'Robbert Harms'
__date__ = '2018-08-14'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class FSLSamplingRoutine(AbstractRWMSampler):

    def __init__(self, ll_func, log_prior_func, x0, proposal_stds, batch_size=50, min_val=1e-15, max_val=1e3, **kwargs):
        r"""An acceptance rate scaling algorithm found in a Neuroscience package called FSL.

        This scaling algorithm scales the std. by :math:`\sqrt(a/(n - a))` where a is the number of accepted samples
        in the last batch and n is the batch size. Its goal is to balance the acceptance rate at 0.5.

        Since this method never ceases the adaptation of the standard deviations, it theoretically loses
        ergodicity of the chain [1, 2].

        Args:
            ll_func (mot.lib.cl_function.CLFunction): The log-likelihood function. See parent docs.
            log_prior_func (mot.lib.cl_function.CLFunction): The log-prior function. See parent docs.
            x0 (ndarray): the starting positions for the sampler. Should be a two dimensional matrix
                with for every modeling instance (first dimension) and every parameter (second dimension) a value.
            proposal_stds (ndarray): for every parameter and every modeling instance an initial proposal std.
            batch_size (int): the size of the batches inbetween which we update the parameters
            min_val (float): the minimum value the standard deviation can take
            max_val (float): the maximum value the standard deviation can take

        References:
            [1] Roberts GO, Rosenthal JS. Examples of adaptive MCMC. J Comput Graph Stat. 2009;18(2):349-367.
                doi:10.1198/jcgs.2009.06134.
            [2] Roberts GO, Rosenthal JS. Coupling and ergodicity of adaptive Markov chain Monte Carlo algorithms.
                J Appl Probab. 2007;44(March 2005):458-475. doi:10.1239/jap/1183667414.
        """
        super().__init__(ll_func, log_prior_func, x0, proposal_stds, **kwargs)
        self._batch_size = batch_size
        self._min_val = min_val
        self._max_val = max_val
        self._acceptance_counter = np.zeros((self._nmr_problems, self._nmr_params), dtype=np.uint64, order='C')

    def _get_mcmc_method_kernel_data_elements(self):
        kernel_data = super()._get_mcmc_method_kernel_data_elements()
        kernel_data.update({
            'acceptance_counter': Array(self._acceptance_counter, mode='rw', ensure_zero_copy=True)
        })
        return kernel_data

    def _at_acceptance_callback_c_func(self):
        return '''
            void _sampleAccepted(_mcmc_method_data* method_data, ulong current_iteration, uint parameter_ind){
                method_data->acceptance_counter[parameter_ind]++;
            }
        '''

    def _get_proposal_update_function(self, nmr_samples, thinning, return_output):
        kernel_source = '''
            void _updateProposalState(_mcmc_method_data* method_data, ulong current_iteration,
                                      local mot_float_type* current_position){    
                if(current_iteration > 0 && current_iteration % ''' + str(self._batch_size) + ''' == 0){
                    for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                        
                        method_data->proposal_stds[k] *= sqrt(
                            ((mot_float_type)method_data->acceptance_counter[k]) /
                            (''' + str(self._batch_size) + ''' - method_data->acceptance_counter[k]));
                        
                        method_data->proposal_stds[k] = clamp(method_data->proposal_stds[k], 
                                                        (mot_float_type)''' + str(self._min_val) + ''', 
                                                        (mot_float_type)''' + str(self._max_val) + ''');

                        method_data->acceptance_counter[k] = 0;
                    }
                }             
            }
        '''
        return kernel_source



import numpy as np
from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = '2017-07-19'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class ActiveAx(CompositeModelTemplate):

    description = '''
        The ActiveAx model for use in in-vivo measurements.

        This model has a CSF compartment and white matter intrinsic diffusivity is fixed at 1.7E-9 m^2/s and 
        CSF diffusivity at 3.0E-9 m^2/s. (In CAMINO this model is listed as "MMWMD_INVIVO").
    '''
    model_expression = '''
        S0 * ((Weight(w_ic) * CylinderGPD) +
              (Weight(w_ec) * Zeppelin) +
              (Weight(w_csf) * Ball))
    '''
    inits = {
        'w_csf.w': 0.01
    }
    fixes = {'CylinderGPD.d': 1.7e-9,
             'Ball.d': 3.0e-9,
             'Zeppelin.d': 'CylinderGPD.d',
             'Zeppelin.dperp0': 'Zeppelin.d * (w_ec.w / (w_ec.w + w_ic.w))',
             'Zeppelin.theta': 'CylinderGPD.theta',
             'Zeppelin.phi': 'CylinderGPD.phi'}
    extra_optimization_maps = [
        lambda d: {'AxonDensityIndex': (4 * (d['w_ic.w'] / (d['w_ec.w'] + d['w_ic.w'])))
                                       / (np.pi * (2 * d['CylinderGPD.R']) ** 2)}
    ]


class ActiveAx_ExVivo(CompositeModelTemplate):

    description = '''
        The ActiveAx model for use in ex-vivo measurements.

        This model has all the compartments as described in Alexander et al NIMG 2010. The white matter intrinsic
        diffusivity is fixed at 0.6e-9 m^2/s and CSF diffusivity is fixed at 2.0e-9 m^2/s. 
        (In CAMINO this model is listed as "MMWMD_Fixed").
    '''
    model_expression = '''
        S0 * ((Weight(w_ic) * CylinderGPD) +
              (Weight(w_ec) * Zeppelin) +
              (Weight(w_csf) * Ball) + 
              (Weight(w_stat) * Dot))
    '''
    inits = {
        'w_csf.w': 0.01,
        'w_stat.w': 0.01
    }
    fixes = {
        'CylinderGPD.d': 0.6e-9,
        'Ball.d': 2.0e-9,
        'Zeppelin.d': 'CylinderGPD.d',
        'Zeppelin.dperp0': 'Zeppelin.d * (w_ec.w / (w_ec.w + w_ic.w))',
        'Zeppelin.theta': 'CylinderGPD.theta',
        'Zeppelin.phi': 'CylinderGPD.phi'
    }
    extra_optimization_maps = [
        lambda d: {'AxonDensityIndex': (4 * (d['w_ic.w'] / (d['w_ec.w'] + d['w_ic.w'])))
                                       / (np.pi * (2 * d['CylinderGPD.R']) ** 2)}
    ]


class TimeDependentActiveAx(CompositeModelTemplate):

    description = 'Fits the extra-axonal time dependent ActiveAx model (De Santis 2016).'
    model_expression = '''
        S0 * ExpT1DecTM_simple * ( (Weight(w_ic) * CylinderGPD) + 
                                   (Weight(w_ec) * TimeDependentZeppelin)
                                  )
    '''
    fixes = {'TimeDependentZeppelin.d': 'CylinderGPD.d',
             'TimeDependentZeppelin.theta': 'CylinderGPD.theta',
             'TimeDependentZeppelin.phi': 'CylinderGPD.phi',
             }
    extra_optimization_maps = [
        lambda d: {'AxonDensityIndex': (4 * (d['w_ic.w'] / (d['w_ec.w'] + d['w_ic.w'])))
                                       / (np.pi * (2 * d['CylinderGPD.R']) ** 2)}
    ]



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Launches the MDT maps visualizer."""
import argparse
import os
import textwrap
from argcomplete.completers import FilesCompleter
from mdt import view_maps
from mdt.visualization.maps.base import SimpleDataInfo
from mdt.lib.shell_utils import BasicShellApplication

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GUI(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('items', metavar='items', type=str, nargs='*', help='the directory or file(s)',
                            default=None).completer = FilesCompleter()

        parser.add_argument('-c', '--config', type=str,
                            help='Use the given initial configuration').completer = \
            FilesCompleter(['conf'], directories=False)

        parser.add_argument('-m', '--maximize', action='store_true', help="Maximize the shown window")
        parser.add_argument('--to-file', type=str, help="If set export the figure to the given filename")

        parser.add_argument('--width', type=int, help="The width of the output file when --to-file is set")
        parser.add_argument('--height', type=int, help="The height of the output file when --to-file is set")
        parser.add_argument('--dpi', type=int, help="The dpi of the output file when --to-file is set")

        return parser

    def run(self, args, extra_args):
        if args.items:
            items = []
            for path in args.items:
                items.append(os.path.realpath(path))
            data = SimpleDataInfo.from_paths(items)
        else:
            data = SimpleDataInfo.from_paths([os.getcwd()])

        to_file = None
        if args.to_file:
            to_file = os.path.realpath(args.to_file)

        config = None
        if args.config:
            filename = os.path.realpath(args.config)
            if os.path.exists(filename):
                with open(filename, 'r') as f:
                    config = f.read()

        figure_options = {}
        if to_file:
            if args.width:
                figure_options.update({'width': args.width})
            if args.height:
                figure_options.update({'height': args.height})
            if args.dpi:
                figure_options.update({'dpi': args.dpi})

            view_maps(data, config, figure_options=figure_options, save_filename=to_file)
        else:
            view_maps(data, config, show_maximized=args.maximize)


def get_doc_arg_parser():
    return GUI().get_documentation_arg_parser()


if __name__ == '__main__':
    GUI().start()



import copy
import os

from PyQt5.QtCore import pyqtSlot, Qt, QPoint, QUrl
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import QWidget, QAbstractItemView, QMenu, QWidgetAction, QLabel

from mdt.gui.maps_visualizer.actions import SetDimension, SetSliceIndex, SetVolumeIndex, SetColormap, SetRotate, \
    SetZoom, SetShowAxis, SetColorBarNmrTicks, SetMapsToShow, SetFont, SetInterpolation, SetFlipud, SetPlotTitle, \
    SetGeneralMask, NewDataAction, SetShowPlotColorbars, SetColorbarLocation, SetShowPlotTitles
from mdt.gui.maps_visualizer.base import DataConfigModel
from mdt.gui.maps_visualizer.design.ui_TabGeneral import Ui_TabGeneral
from mdt.gui.utils import blocked_signals, TimedUpdate, split_long_path_elements
from mdt.visualization.maps.base import Zoom, Point2d, DataInfo, Font, MapPlotConfig

__author__ = 'Robbert Harms'
__date__ = "2016-09-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class TabGeneral(QWidget, Ui_TabGeneral):

    def __init__(self, controller, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self._controller = controller
        self._controller.model_updated.connect(self.update_model)

        self._previous_model = None
        current_model = self._controller.get_model()

        self.general_display_order.setDragDropMode(QAbstractItemView.InternalMove)
        self.general_display_order.setSelectionMode(QAbstractItemView.SingleSelection)

        self.general_colormap.addItems(current_model.get_config().get_available_colormaps())
        self.general_rotate.addItems(['0', '90', '180', '270'])
        self.general_rotate.setCurrentText(str(current_model.get_config().rotate))

        self.general_DisplayOrder.set_collapse(True)
        self.general_Miscellaneous.set_collapse(True)
        self.general_Zoom.set_collapse(True)
        self.general_Font.set_collapse(True)
        self.general_Colorbar.set_collapse(True)

        self.general_dimension.valueChanged.connect(lambda v: self._controller.apply_action(SetDimension(v)))
        self.general_slice_index.valueChanged.connect(lambda v: self._controller.apply_action(SetSliceIndex(v)))
        self.general_volume_index.valueChanged.connect(lambda v: self._controller.apply_action(SetVolumeIndex(v)))
        self.general_colormap.currentIndexChanged.connect(
            lambda i: self._controller.apply_action(SetColormap(self.general_colormap.itemText(i))))
        self.general_rotate.currentIndexChanged.connect(
            lambda i: self._controller.apply_action(SetRotate(int(self.general_rotate.itemText(i)))))

        self._map_selection_timer = TimedUpdate(self._update_maps_to_show)
        self.general_map_selection.itemSelectionChanged.connect(
            lambda: self._map_selection_timer.add_delayed_callback(500))
        self.general_map_selection.setContextMenuPolicy(Qt.CustomContextMenu)
        self.general_map_selection.customContextMenuRequested.connect(self.select_maps_context_menu)

        self.general_deselect_all_maps.clicked.connect(self._deleselect_all_maps)
        self.general_invert_map_selection.clicked.connect(self._invert_map_selection)

        self.general_zoom_x_0.valueChanged.connect(self._update_zoom)
        self.general_zoom_x_1.valueChanged.connect(self._update_zoom)
        self.general_zoom_y_0.valueChanged.connect(self._update_zoom)
        self.general_zoom_y_1.valueChanged.connect(self._update_zoom)

        self.plot_title.textEdited.connect(lambda txt: self._controller.apply_action(SetPlotTitle(txt)))

        self.general_zoom_reset.clicked.connect(lambda: self._controller.apply_action(SetZoom(Zoom.no_zoom())))
        self.general_zoom_fit.clicked.connect(self._zoom_fit)

        self.general_display_order.items_reordered.connect(self._reorder_maps)
        self.general_show_axis.clicked.connect(lambda: self._controller.apply_action(
            SetShowAxis(self.general_show_axis.isChecked())))

        self.general_font_family.addItems(Font.font_names())
        self.general_font_family.currentTextChanged.connect(
            lambda v: self._controller.apply_action(SetFont(
                self._controller.get_model().get_config().font.get_updated(family=v))))

        self.general_font_size.valueChanged.connect(
            lambda: self._controller.apply_action(SetFont(
                self._controller.get_model().get_config().font.get_updated(size=self.general_font_size.value()))))

        self.general_interpolation.addItems(current_model.get_config().get_available_interpolations())
        self.general_interpolation.currentTextChanged.connect(
            lambda v: self._controller.apply_action(SetInterpolation(v)))

        self.general_flipud.clicked.connect(lambda: self._controller.apply_action(
            SetFlipud(self.general_flipud.isChecked())))

        self.general_show_plot_titles.clicked.connect(lambda: self._controller.apply_action(
            SetShowPlotTitles(self.general_show_plot_titles.isChecked())))

        self.mask_name.currentIndexChanged.connect(self._update_mask_name)

        self.general_colorbar_nmr_ticks.valueChanged.connect(
            lambda v: self._controller.apply_action(SetColorBarNmrTicks(v)))
        self.general_show_colorbar.clicked.connect(lambda: self._controller.apply_action(
            SetShowPlotColorbars(self.general_show_colorbar.isChecked())))
        self.general_colorbar_location.currentTextChanged.connect(
            lambda v: self._controller.apply_action(SetColorbarLocation(v.lower())))

    @pyqtSlot(DataConfigModel)
    def update_model(self, model):
        data = model.get_data()
        config = model.get_config()

        if not self._previous_model or model.get_data() != self._previous_model.get_data():
            self._previous_model = model
            self._update_data(data)

        self._update_config(data, config)

    def select_maps_context_menu(self, position):
        global_position = self.general_map_selection.mapToGlobal(position)

        def get_header_action(parent, map_name):
            label = QLabel(map_name)

            font = label.font()
            font.setBold(True)
            label.setFont(font)
            label.setStyleSheet('color: black; margin:5px; margin-left: 15px;')

            action = QWidgetAction(parent)
            action.setDisabled(True)
            action.setDefaultWidget(label)

            return action

        if self.general_map_selection.count():
            row = self.general_map_selection.indexAt(position)
            if row:
                element = self.general_map_selection.item(row.row())
                if element:
                    map_name = element.data(Qt.UserRole)
                    file_path = self._controller.get_model().get_data().get_file_path(map_name)

                    menu = QMenu()
                    menu.addAction(get_header_action(menu, map_name))
                    menu.addSeparator()

                    show_in_folder = menu.addAction('&Show in folder', lambda:
                        QDesktopServices.openUrl(QUrl.fromLocalFile(os.path.dirname(file_path))))
                    if file_path is None:
                        show_in_folder.setEnabled(False)

                    menu.addAction('Use as &mask', lambda: self._controller.apply_action(SetGeneralMask(map_name)))
                    menu.addAction('R&emove', lambda: self._controller.apply_action(
                        NewDataAction(self._controller.get_model().get_data().get_updated(removals=[map_name]))))
                    menu.exec(global_position)

    @pyqtSlot()
    def _reorder_maps(self):
        items = [self.general_display_order.item(ind) for ind in range(self.general_display_order.count())]
        map_names = [item.data(Qt.UserRole) for item in items]
        self._controller.apply_action(SetMapsToShow(map_names))

    @pyqtSlot()
    def _update_maps_to_show(self):
        current_model = self._controller.get_model()
        map_names = copy.copy(current_model.get_config().maps_to_show)

        for item in [self.general_map_selection.item(ind) for ind in range(self.general_map_selection.count())]:
            map_name = item.data(Qt.UserRole)

            if item.isSelected():
                if map_name not in map_names:
                    self._insert_alphabetically(map_name, map_names)
            else:
                if map_name in map_names:
                    map_names.remove(map_name)

        self._controller.apply_action(SetMapsToShow(map_names))

    @pyqtSlot()
    def _deleselect_all_maps(self):
        self._controller.apply_action(SetMapsToShow([]))

    @pyqtSlot()
    def _invert_map_selection(self):
        current_model = self._controller.get_model()
        self._controller.apply_action(SetMapsToShow(
            list(set(current_model.get_data().get_map_names()).difference(
                set(current_model.get_config().maps_to_show)))))

    @pyqtSlot()
    def _zoom_fit(self):
        current_model = self._controller.get_model()
        data_info = current_model.get_data()
        config = current_model.get_config()

        def add_padding(bounding_box, max_x, max_y):
            bounding_box[0].x = max(bounding_box[0].x - 1, 0)
            bounding_box[0].y = max(bounding_box[0].y - 1, 0)

            bounding_box[1].y = min(bounding_box[1].y + 2, max_y)
            bounding_box[1].x = min(bounding_box[1].x + 2, max_x)

            return bounding_box

        if config.maps_to_show or len(data_info.get_map_names()):
            bounding_box = data_info.get_bounding_box(config.dimension, config.slice_index,
                                                      config.volume_index, config.rotate, config.maps_to_show)

            max_y = data_info.get_max_y_index(config.dimension, rotate=config.rotate, map_names=config.maps_to_show)
            max_x = data_info.get_max_x_index(config.dimension, rotate=config.rotate, map_names=config.maps_to_show)

            if not config.flipud:
                # Since the renderer plots with a left top coordinate system,
                # we need to flip the y coordinates upside down by default.
                tmp = max_y - bounding_box[0].y
                bounding_box[0].y = max_y - bounding_box[1].y
                bounding_box[1].y = tmp

            bounding_box = add_padding(bounding_box, max_x, max_y)

            self._controller.apply_action(SetZoom(Zoom(*bounding_box)))

    @pyqtSlot()
    def _update_zoom(self):
        np0x, np0y = self.general_zoom_x_0.value(), self.general_zoom_y_0.value()
        np1x, np1y = self.general_zoom_x_1.value(), self.general_zoom_y_1.value()

        if np0x > np1x:
            np1x = np0x
        if np0y > np1y:
            np1y = np0y

        self._controller.apply_action(SetZoom(Zoom.from_coords(np0x, np0y, np1x, np1y)))

    @staticmethod
    def _insert_alphabetically(new_item, item_list):
        for ind, item in enumerate(item_list):
            if item > new_item:
                item_list.insert(ind, new_item)
                return
        item_list.append(new_item)

    @pyqtSlot(int)
    def _update_mask_name(self, index):
        if index == 0:
            self._controller.apply_action(SetGeneralMask(None))
        else:
            self._controller.apply_action(SetGeneralMask(self.mask_name.itemText(index)))

    def _update_data(self, data_info):
        sorted_keys = list(sorted(data_info.get_map_names()))

        if len(data_info.get_map_names()):
            self.general_info_nmr_maps.setText(str(len(data_info.get_map_names())))
        else:
            self.general_info_nmr_maps.setText('0')

        with blocked_signals(self.general_map_selection):
            self.general_map_selection.clear()
            self.general_map_selection.addItems(sorted_keys)
            for index, map_name in enumerate(sorted_keys):
                item = self.general_map_selection.item(index)
                item.setData(Qt.UserRole, map_name)

        with blocked_signals(self.mask_name):
            self.mask_name.clear()
            self.mask_name.insertItem(0, '-- None --')
            self.mask_name.insertItems(1, sorted_keys)

    def _update_config(self, data, config):
        map_names = config.maps_to_show

        with blocked_signals(self.general_dimension):
            try:
                max_dimension = data.get_max_dimension(map_names)
                self.general_dimension.setMaximum(max_dimension)
                self.maximumDimension.setText(str(max_dimension))
            except ValueError:
                self.general_dimension.setMaximum(0)
                self.maximumDimension.setText(str(0))
            self.general_dimension.setValue(config.dimension)

        with blocked_signals(self.general_slice_index):
            try:
                max_slice = data.get_max_slice_index(config.dimension, map_names)
                self.general_slice_index.setMaximum(max_slice)
                self.maximumIndex.setText(str(max_slice))
            except ValueError:
                self.general_slice_index.setMaximum(0)
                self.maximumIndex.setText(str(0))
            self.general_slice_index.setValue(config.slice_index)

        with blocked_signals(self.general_volume_index):
            try:
                max_volume = data.get_max_volume_index(map_names)
                self.general_volume_index.setMaximum(max_volume)
                self.maximumVolume.setText(str(max_volume))
            except ValueError:
                self.general_volume_index.setMaximum(0)
                self.maximumVolume.setText(str(0))
            self.general_volume_index.setValue(config.volume_index)

        with blocked_signals(self.general_colormap):
            self.general_colormap.setCurrentText(config.colormap)

        with blocked_signals(self.general_rotate):
            self.general_rotate.setCurrentText(str(config.rotate))

        if self.general_map_selection.count():
            for map_name, map_config in config.map_plot_options.items():
                if map_config.title:
                    index = list(sorted(data.get_map_names())).index(map_name)
                    item = self.general_map_selection.item(index)
                    item.setData(Qt.DisplayRole, map_name + ' (' + map_config.title + ')')

            self.general_map_selection.blockSignals(True)
            for index, map_name in enumerate(list(sorted(data.get_map_names()))):
                item = self.general_map_selection.item(index)
                if item:
                    item.setSelected(map_name in map_names)
            self.general_map_selection.blockSignals(False)

        try:
            max_x = data.get_max_x_index(config.dimension, config.rotate, map_names)
            max_y = data.get_max_y_index(config.dimension, config.rotate, map_names)

            with blocked_signals(self.general_zoom_x_0, self.general_zoom_x_1,
                                 self.general_zoom_y_0, self.general_zoom_y_1):
                self.general_zoom_x_0.setMaximum(max_x)
                self.general_zoom_x_0.setValue(config.zoom.p0.x)

                self.general_zoom_x_1.setMaximum(max_x)
                self.general_zoom_x_1.setMinimum(config.zoom.p0.x)
                self.general_zoom_x_1.setValue(config.zoom.p1.x)

                self.general_zoom_y_0.setMaximum(max_y)
                self.general_zoom_y_0.setValue(config.zoom.p0.y)

                self.general_zoom_y_1.setMaximum(max_y)
                self.general_zoom_y_1.setMinimum(config.zoom.p0.y)
                self.general_zoom_y_1.setValue(config.zoom.p1.y)

                if config.zoom.p0.x == 0 and config.zoom.p1.x == 0:
                    self.general_zoom_x_1.setValue(max_x)

                if config.zoom.p0.y == 0 and config.zoom.p1.y == 0:
                    self.general_zoom_y_1.setValue(max_y)
        except ValueError:
            pass

        with blocked_signals(self.plot_title):
            self.plot_title.setText(config.title)

        with blocked_signals(self.general_display_order):
            self.general_display_order.clear()
            self.general_display_order.addItems(map_names)

            for index, map_name in enumerate(map_names):
                item = self.general_display_order.item(index)
                item.setData(Qt.UserRole, map_name)

                if map_name in config.map_plot_options and config.map_plot_options[map_name].title:
                    title = config.map_plot_options[map_name].title
                    item.setData(Qt.DisplayRole, map_name + ' (' + title + ')')

        with blocked_signals(self.general_show_axis):
            self.general_show_axis.setChecked(config.show_axis)

        with blocked_signals(self.general_colorbar_nmr_ticks):
            self.general_colorbar_nmr_ticks.setValue(config.colorbar_settings.get_preferred('nmr_ticks'))

        with blocked_signals(self.general_show_colorbar):
            self.general_show_colorbar.setChecked(config.colorbar_settings.get_preferred('visible'))

        with blocked_signals(self.general_colorbar_location):
            self.general_colorbar_location.setCurrentText(config.colorbar_settings.get_preferred('location').title())

        with blocked_signals(self.general_show_plot_titles):
            self.general_show_plot_titles.setChecked(config.show_titles)

        with blocked_signals(self.general_font_family):
            self.general_font_family.setCurrentText(config.font.family)

        with blocked_signals(self.general_font_size):
            self.general_font_size.setValue(config.font.size)

        with blocked_signals(self.general_interpolation):
            self.general_interpolation.setCurrentText(config.interpolation)

        with blocked_signals(self.general_flipud):
            self.general_flipud.setChecked(config.flipud)

        with blocked_signals(self.mask_name):
            if config.mask_name and config.mask_name in data.get_map_names():
                for ind in range(self.mask_name.count()):
                    if self.mask_name.itemText(ind) == config.mask_name:
                        self.mask_name.setCurrentIndex(ind)
                        break
            else:
                self.mask_name.setCurrentIndex(0)



import glob
import gzip
import os
import copy
import nibabel as nib
import numpy as np
import shutil

from mdt.lib.deferred_mappings import DeferredActionDict

__author__ = 'Robbert Harms'
__date__ = "2014-08-28"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def load_nifti(nifti_volume):
    """Load and return a nifti file.

    This will apply path resolution if a filename without extension is given. See the function
    :func:`nifti_filepath_resolution` for details.

    Args:
        nifti_volume (string): The filename of the volume to use.

    Returns:
        :class:`nibabel.nifti2.Nifti2Image`
    """
    path = nifti_filepath_resolution(nifti_volume)
    return nifti_info_decorate_nibabel_image(nib.load(path))


def load_all_niftis(directory, map_names=None):
    """Loads all niftis in the given directory as nibabel nifti files.

    This does not load the data directly, it loads the niftis in a dictionary. To get a direct handle to the image
    data use the function :func:`get_all_nifti_data`.

    If ``map_names`` is given we will only load the given maps. Else, we will load all .nii and .nii.gz files.
    The map name is the filename of a nifti without the extension.

    In the case both an .nii and a .nii.gz with the same name exists we will load the .nii as the main map
    and the .nii.gz with its extension.

    Args:
        directory (str): the directory from which we want to load the niftis
        map_names (list of str): the names of the maps we want to use. If given, we only use and return these maps.

    Returns:
        dict: A dictionary with the loaded nibabel proxies (see :func:`load_nifti`).
            The keys of the dictionary are the filenames without the extension of the .nii(.gz) files
            in the given directory.
    """
    maps_paths = {}

    for path, map_name, extension in yield_nifti_info(directory):
        if not map_names or map_name in map_names:
            if map_name in maps_paths:
                map_name += extension
            maps_paths.update({map_name: path})

    return {k: load_nifti(v) for k, v in maps_paths.items()}


def get_all_nifti_data(directory, map_names=None, deferred=True):
    """Get the data of all the nifti volumes in the given directory.

    If map_names is given we will only load the given map names. Else, we load all .nii and .nii.gz files in the
    given directory.

    Args:
        directory (str): the directory from which we want to read a number of maps
        map_names (list of str): the names of the maps we want to use. If given, we only use and return these maps.
        deferred (boolean): if True we return an deferred loading dictionary instead of a dictionary with the values
            loaded as arrays.

    Returns:
        dict: A dictionary with the volumes. The keys of the dictionary are the filenames
            without the extension of the .nii(.gz) files in the given directory.
    """
    proxies = load_all_niftis(directory, map_names=map_names)
    if deferred:
        return DeferredActionDict(lambda _, item: item.get_data(), proxies)
    else:
        return {k: v.get_data() for k, v in proxies.items()}


def write_nifti(data, output_fname, header=None, affine=None, use_data_dtype=True, **kwargs):
    """Write data to a nifti file.

    This will write the output directory if it does not exist yet.

    Args:
        data (ndarray): the data to write to that nifti file
        output_fname (str): the name of the resulting nifti file, this function will append .nii.gz if no
            suitable extension is given.
        header (nibabel header): the nibabel header to use as header for the nifti file. If None we will use
            a default header.
        affine (ndarray): the affine transformation matrix
        use_data_dtype (boolean): if we want to use the dtype from the data instead of that from the header
            when saving the nifti.
        **kwargs: other arguments to Nifti2Image from NiBabel
    """
    if header is None:
        header = nib.nifti2.Nifti2Header()

    if use_data_dtype:
        header = copy.deepcopy(header)
        dtype = data.dtype
        if data.dtype == np.bool:
            dtype = np.char
        try:
            header.set_data_dtype(dtype)
        except nib.spatialimages.HeaderDataError:
            pass

    if not (output_fname.endswith('.nii.gz') or output_fname.endswith('.nii')):
        output_fname += '.nii.gz'

    if not os.path.exists(os.path.dirname(output_fname)):
        os.makedirs(os.path.dirname(output_fname))

    if isinstance(header, nib.nifti2.Nifti2Header):
        format = nib.Nifti2Image
    else:
        format = nib.Nifti1Image

    format(data, affine, header=header, **kwargs).to_filename(output_fname)


def write_all_as_nifti(volumes, directory, nifti_header=None, overwrite_volumes=True, gzip=True):
    """Write a number of volume maps to the specific directory.

    Args:
        volumes (dict): the volume maps (in 3d) with the results we want to write.
            The filenames are generated using the keys in the given volumes
        directory (str): the directory to write to
        nifti_header: the nifti header to use for each of the volumes.
        overwrite_volumes (boolean): defaults to True, if we want to overwrite the volumes if they exists
        gzip (boolean): if True we write the files as .nii.gz, if False we write the files as .nii
    """
    for key, volume in volumes.items():
        extension = '.nii'
        if gzip:
            extension += '.gz'
        filename = key + extension

        full_filename = os.path.abspath(os.path.join(directory, filename))

        if os.path.exists(full_filename):
            if overwrite_volumes:
                os.remove(full_filename)
                write_nifti(volume, full_filename, header=nifti_header)
        else:
            write_nifti(volume, full_filename, header=nifti_header)


def nifti_filepath_resolution(file_path):
    """Tries to resolve the filename to a nifti based on only the filename.

    For example, this resolves the path: ``/tmp/mask`` to:

        - ``/tmp/mask`` if exists
        - ``/tmp/mask.nii`` if exist
        - ``/tmp/mask.nii.gz`` if exists

    Hence, the lookup order is: ``path``, ``path.nii``, ``path.nii.gz``

    If a file with an extension is given we will do no further resolving and return the path as is.

    Args:
        file_path (str): the path to the nifti file, can be without extension.

    Returns:
        str: the file path we resolved to the final file.

    Raises:
        ValueError: if no nifti file could be found
    """
    if file_path[:-len('.nii')] == '.nii' or file_path[:-len('.nii.gz')] == '.nii.gz':
        return file_path

    if os.path.isfile(file_path):
        return file_path
    elif os.path.isfile(file_path + '.nii'):
        return file_path + '.nii'
    elif os.path.isfile(file_path + '.nii.gz'):
        return file_path + '.nii.gz'
    raise ValueError('No nifti file could be found using the path {}'.format(file_path))


def yield_nifti_info(directory):
    """Get information about the nifti volumes in the given directory.

    Args:
        directory (str): the directory to get the names of the available maps from

    Yields:
        tuple: (path, map_name, extension) for every map found
    """
    for extension in ('.nii', '.nii.gz'):
        for f in glob.glob(os.path.join(directory, '*' + extension)):
            yield f, os.path.basename(f)[0:-len(extension)], extension


def is_nifti_file(file_name):
    """Check if the given file is a nifti file.

    This only checks if the extension of the given file ends with .nii or with .nii.gz

    Args:
        file_name (str): the name of the file

    Returns:
        boolean: true if the file looks like a nifti file, false otherwise
    """
    return file_name.endswith('.nii') or file_name.endswith('.nii.gz')


def unzip_nifti(input_filename, output_filename):
    """Unzips the given nifti file.

    This will create the output directories if they do not yet exist.

    Args:
        input_filename (str): the nifti file we would like to unzip. Should have the extension ``.gz``.
        output_filename (str): the location for the output file. Should have the extension ``.nii``.

    Raises:
        ValueError: if the extensions of either the input or output filename are not correct.
    """
    if not os.path.exists(os.path.dirname(output_filename)):
        os.makedirs(os.path.dirname(output_filename))

    if not input_filename.rstrip().endswith('.gz') or not output_filename.rstrip().endswith('.nii'):
        raise ValueError('The input filename should have extension ".gz" and the '
                         'output filename should have extension ".nii".')

    with gzip.open(input_filename, 'rb') as f_in, open(output_filename, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)


class NiftiInfo(object):

    def __init__(self, header=None, filepath=None):
        """A nifti information object to store meta data alongside an array.

        Args:
            header: the nibabel nifti header
            filepath (str): the on-disk filepath of the corresponding nifti file
        """
        self.header = header
        self.filepath = filepath


class NiftiInfoDecorated(object):
    """The additional type of an array after it has been subclassed by :func:`nifti_info_decorate_array`.

    This subclass can be used to check if an array has nifti info attached to it.
    """

    @property
    def nifti_info(self):
        """Get the nifti information attached to the subclass.

        Returns:
            NiftiInfo: the nifti information object
        """
        raise NotImplementedError()


def nifti_info_decorate_array(array, nifti_info=None):
    """Decorate the provided numpy array with nifti information.

    This can be used to ensure that an array is of additional subclass :class:`NiftiInfoDecorated`.

    Args:
        array (ndarray): the numpy array to decoreate
        nifti_info (NiftiInfo): the nifti info to attach to the array
    """
    class NiftiInfoDecoratedArray(type(array), NiftiInfoDecorated):
        def __new__(cls, input_array, nifti_info=None):
            """Decorate an existing input array with some additional about a nifti file.

            This is typically used to store the original nifti header and the filepath as an special attribute of
            the array. Please note that this metadata does not survive subarrays and views of this array.

            Args:
                input_array (ndarray): the array we decorate
                nifti_info (NiftiInfo): a nifti information object

            Returns:
                ndarray: the decorated array
            """
            obj = input_array.view(cls)
            obj._nifti_info = nifti_info or NiftiInfo()
            return obj

        @property
        def nifti_info(self):
            return self._nifti_info

        def __array_finalize__(self, obj):
            if obj is None:
                return
            self._nifti_info = getattr(obj, '_nifti_info', None)

    return NiftiInfoDecoratedArray(array, nifti_info)


def nifti_info_decorate_nibabel_image(nifti_obj):
    """Decorate the nibabel image container such that the ``get_data`` method returns a NiftiInfoDecorated ndarray.

    Args:
        nifti_obj: a nibabel nifti object
    """
    original_function = nifti_obj.get_data

    def get_data(self, *args, **kwargs):
        data = original_function(*args, **kwargs)
        return nifti_info_decorate_array(data, NiftiInfo(header=self.get_header(), filepath=self.get_filename()))

    nifti_obj.get_data = get_data.__get__(nifti_obj, type(nifti_obj))
    return nifti_obj



import numpy as np
import collections
from mdt.lib.components import get_model
from mdt.lib.nifti import get_all_nifti_data
from mdt.utils import create_roi, restore_volumes, MockMRIInputData
from mot.lib.cl_function import SimpleCLFunction
from mot.lib.kernel_data import Array, Zeros

__author__ = 'Robbert Harms'
__date__ = '2017-05-29'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


def create_signal_estimates(model, input_data, parameters):
    """Create the signals estimates for your estimated model parameters.

    This function is typically used to obtain signal estimates from optimization results.

    This function evaluates the model as it is in the model fitting and sample. That is, this method includes
    the gradient deviations (if set in the input data) and loads all static and fixed parameters maps.

    Args:
        model (str or model): the model or the name of the model to use for estimating the signals
        input_data (mdt.utils.MRIInputData): the input data object, we will set this to the model
        parameters (str or dict): either a directory file name or a dictionary containing optimization results
            Each element is assumed to be a 4d volume with the voxels we are using for the simulations.

    Returns:
        ndarray: the 4d array with the signal estimates per voxel
    """
    if isinstance(model, str):
        model = get_model(model)()

    model.set_input_data(input_data)
    build_model = model.build()

    if isinstance(parameters, str):
        parameters = get_all_nifti_data(parameters)

    parameters = create_roi(parameters, input_data.mask)
    parameters = model.param_dict_to_array(parameters)

    kernel_data = {'data': build_model.get_kernel_data(),
                   'parameters': Array(parameters, ctype='mot_float_type'),
                   'estimates': Zeros((parameters.shape[0], build_model.get_nmr_observations()), 'mot_float_type')}

    _get_simulate_function(build_model).evaluate(kernel_data, parameters.shape[0])
    results = kernel_data['estimates'].get_data()

    return restore_volumes(results, input_data.mask)


def simulate_signals(model, protocol, parameters):
    """Estimate the signals of a given model for the given combination of protocol and parameters.

    In contrast to the function :func:`create_signal_estimates`, this function does not incorporate the gradient
    deviations. Furthermore, this function expects a two dimensional list of parameters and this function will
    simply evaluate the model for each set of parameters.

    Args:
        model (str or model): the model or the name of the model to use for estimating the signals
        protocol (mdt.protocols.Protocol): the protocol we will use for the signal simulation
        parameters (dict or ndarray): the parameters for which to simulate the signal. It can either be a matrix with
            for every row every model parameter, or a dictionary with for every parameter a 1d array.

    Returns:
        ndarray: a 2d array with for every parameter combination the simulated model signal
    """
    if isinstance(model, str):
        model = get_model(model)()

    model.set_input_data(MockMRIInputData(protocol=protocol))
    build_model = model.build()

    if isinstance(parameters, collections.Mapping):
        parameters = model.param_dict_to_array(parameters)

    nmr_problems = parameters.shape[0]

    kernel_data = {'data': build_model.get_kernel_data(),
                   'parameters': Array(parameters, ctype='mot_float_type'),
                   'estimates': Zeros((nmr_problems, build_model.get_nmr_observations()), 'mot_float_type')
                   }

    _get_simulate_function(build_model).evaluate(kernel_data, nmr_problems)
    return kernel_data['estimates'].get_data()


def add_rician_noise(signals, noise_level, seed=None):
    """Make the given signal Rician distributed.

    To calculate the noise level divide the signal of the unweighted volumes by the SNR you want. For example,
    for a unweighted signal b0=1e4 and a desired SNR of 20, you need an noise level of 1e4/20 = 500.

    Args:
        signals: the signals to make Rician distributed
        noise_level: the level of noise to add. The actual Rician stdev depends on the signal. See ricestat in
            the mathworks library. The noise level can be calculated using b0/SNR.
        seed (int): if given, the seed for the random number generation

    Returns:
        ndarray: make every element of the input signals contain Rician distributed noise.
    """
    random_state = np.random.RandomState(seed)
    x = noise_level * random_state.normal(size=signals.shape) + signals
    y = noise_level * random_state.normal(size=signals.shape)
    return np.sqrt(np.power(x, 2), np.power(y, 2)).astype(signals.dtype)


def _get_simulate_function(model):
    """Get the simulation function.

    This is a wrapper around the model evaluation function. As such, it includes the protocol adaptation callbacks.
    It creates the model signal just as used by the log-likelihood function during model optimization.

    Please note that this function does not include the
    """
    eval_function_info = model.get_model_eval_function()
    return SimpleCLFunction.from_string('''
        void simulate(void* data, local mot_float_type* parameters, global mot_float_type* estimates){
            for(uint i = 0; i < ''' + str(model.get_nmr_observations()) + '''; i++){
                estimates[i] = ''' + eval_function_info.get_cl_function_name() + '''(data, parameters, i);
            }
        }
    ''', dependencies=[eval_function_info])




from mdt.component_templates.compartment_models import CompartmentTemplate

__author__ = 'Robbert Harms'
__date__ = "2018-08-28"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AstroCylinders(CompartmentTemplate):

    parameters = ('b', 'G', 'Delta', 'delta', 'd', 'R')
    dependencies = ['VanGelderenCylinder']
    cl_code = '''
        if(b == 0){
            return 1;
        }
        
        mot_float_type lperp = VanGelderenCylinder(G, Delta, delta, d, R) / (G * G);
        mot_float_type lpar = -(b / (G*G)) * d;

        return (sqrt(M_PI) / (2 * G * sqrt(lperp - lpar)))
                    * exp(G * G * lperp)
                    * erf(G * sqrt(lperp - lpar));
    '''



from collections import Mapping, Sequence

__author__ = 'Robbert Harms'
__date__ = "2016-09-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ConversionSpecification(object):

    def __init__(self):
        """Specifies how the content of an object is to be converted from and to a dictionary."""

    def to_dict(self, obj):
        """Convert the given value to a dictionary.

        Args:
            obj (object): the value to convert to a dictionary

        Returns:
            dict: the resulting converted dictionary
        """

    def from_dict(self, value):
        """Generate a result value from the given dictionary

        Args:
            value (object): the dictionary to convert back to a value

        Returns:
            object: the value represented by the dict.
        """


class OptionalConversionDecorator(ConversionSpecification):

    def __init__(self, conversion_specification):
        """Makes the conversion optional by testing against None.

        If the element to convert is None, we will return None as a conversion. If the element to convert is not None
        we will convert it according to the conversion specified.

        This holds for both to- and from- dict.

        Args:
            conversion_specification (ConversionSpecification): the conversion specification to use if the element
                to convert is not None.
        """
        super().__init__()
        self._conversion_specification = conversion_specification

    def to_dict(self, obj):
        if obj is None:
            return None
        return self._conversion_specification.to_dict(obj)

    def from_dict(self, value):
        if value is None:
            return None
        return self._conversion_specification.from_dict(value)


class SimpleClassConversion(ConversionSpecification):

    def __init__(self, class_type, attribute_conversions):
        """Converts a dictionary to and from the specified class.

        Args:
            class_type (type): the type of class to convert to
            attribute_conversions (List[ConversionSpecification]) the list of conversion specification for the
                attributes
        """
        super().__init__()
        self._class_type = class_type
        self._attribute_conversions = attribute_conversions

    def to_dict(self, obj):
        result_dict = {}
        conversion_info = self._attribute_conversions
        for key, converter in conversion_info.items():
            result_dict[key] = converter.to_dict(getattr(obj, key))
        return result_dict

    def from_dict(self, value):
        init_kwargs = {}
        for key, converter in self._attribute_conversions.items():
            if key in value:
                init_kwargs[key] = converter.from_dict(value[key])
        return self._class_type(**init_kwargs)


class ConvertDictElements(ConversionSpecification):

    def __init__(self, conversion_type):
        """Converts all the elements in the value (a dictionary) using the given conversion type."""
        super().__init__()
        self._conversion_type = conversion_type

    def to_dict(self, obj):
        return {key: self._conversion_type.to_dict(v) for key, v in obj.items()}

    def from_dict(self, value):
        return {key: self._conversion_type.from_dict(v) for key, v in value.items()}


class ConvertListElements(ConversionSpecification):

    def __init__(self, conversion_type):
        """Converts all the elements in the value (a list) using the given conversion type."""
        super().__init__()
        self._conversion_type = conversion_type

    def to_dict(self, obj):
        return [self._conversion_type.to_dict(v) for v in obj]

    def from_dict(self, value):
        return [self._conversion_type.from_dict(v) for v in value]


class ConvertDynamicFromModule(ConversionSpecification):

    def __init__(self, module):
        """Performs dynamic lookup by loading the class from the given module.

        This requires that the class we are dynamically loading has a get_conversion_info() class method that
        returns the conversion specification for that class.

        Args:
            module (module): the python module to use for loading the data from dict
        """
        super().__init__()
        self._module = module

    def to_dict(self, obj):
        return [obj.__class__.__name__, obj.get_conversion_info().to_dict(obj)]

    def from_dict(self, value):
        try:
            cls = getattr(self._module, value[0])
        except AttributeError:
            raise ValueError('The given class "{}" could not be found.'.format(value[0]))

        return cls.get_conversion_info().from_dict(value[1])


class SimpleFunctionConversion(ConversionSpecification):

    def __init__(self, conversion_func=None, allow_null=True, set_null_to_value=None):
        """Performs identity conversion between simple types.

        Args:
            conversion_func (Func): if not None we apply the given function before converting to and from the
                dictionary. Can also be a type like ``int``.
            allow_null (bool): if True we allow None during type casting
            set_null_to_value (obj): the value to set null entries to. If this is None and allow_null is False we
                raise an error.
        """
        super().__init__()
        self._conversion_func = conversion_func
        self._allow_none = allow_null
        self._set_null_to_value = set_null_to_value

    def to_dict(self, obj):
        if obj is None:
            if self._allow_none:
                return None
            else:
                if self._set_null_to_value is not None:
                    return self._set_null_to_value
                raise ValueError('The object is supposed to be not None.')
        else:
            if self._conversion_func:
                return self._conversion_func(obj)
        return obj

    def from_dict(self, value):
        if value is None:
            if self._allow_none:
                return None
            else:
                if self._set_null_to_value is not None:
                    return self._set_null_to_value
                raise ValueError('The object is supposed to be not None.')
        else:
            if self._conversion_func:
                return self._conversion_func(value)
        return value


class SimpleDictConversion(SimpleFunctionConversion):

    def __init__(self, conversion_func=None, allow_null=True, set_null_to_value=None):
        """Converts all the objects in the given dict.

        Args:
            conversion_func (Func): if not None we cast the from_dict value to the given type
            allow_null (bool): if True we allow None during type casting
            set_null_to_value (obj): the value to set null entries to. If this is None and allow_null is False we
                raise an error.
        """
        super().__init__(
            conversion_func=SimpleDictConversion._get_conversion_func(conversion_func),
            allow_null=allow_null,
            set_null_to_value=set_null_to_value)

    @staticmethod
    def _get_conversion_func(user_conversion_func):
        """Wraps the desired type into a Dict[user_conversion_func] function."""
        if user_conversion_func is None:
            return None

        def conversion_wrapper(obj):
            if isinstance(obj, Mapping):
                return {key: user_conversion_func(v) for key, v in obj.items()}
            return obj

        return conversion_wrapper


class SimpleListConversion(SimpleFunctionConversion):

    def __init__(self, conversion_func=None, allow_null=True, set_null_to_value=None):
        """Converts all the objects in the given list.

        Args:
            conversion_func (Func): if not None we cast the from_dict value to the given type
            allow_null (bool): if True we allow None during type casting
            set_null_to_value (obj): the value to set null entries to. If this is None and allow_null is False we
                raise an error.
        """
        super().__init__(
            conversion_func=SimpleListConversion._get_conversion_func(conversion_func),
            allow_null=allow_null,
            set_null_to_value=set_null_to_value)

    @staticmethod
    def _get_conversion_func(desired_type):
        """Wraps the desired type into a dict(desired_type) function."""
        if desired_type is None:
            return None

        def conversion_wrapper(obj):
            if isinstance(obj, Sequence):
                return [desired_type(el) for el in obj]
            return obj

        return conversion_wrapper


class StringConversion(SimpleFunctionConversion):

    def __init__(self, allow_null=True):
        super().__init__(str, allow_null=allow_null)


class IntConversion(SimpleFunctionConversion):

    def __init__(self, allow_null=True):
        super().__init__(int, allow_null=allow_null)


class FloatConversion(SimpleFunctionConversion):

    def __init__(self, allow_null=True):
        super().__init__(float, allow_null=allow_null)


class BooleanConversion(SimpleFunctionConversion):

    def __init__(self, allow_null=True):
        super().__init__(bool, allow_null=allow_null)


class WhiteListConversion(ConversionSpecification):

    def __init__(self, white_list, default):
        """Allow only elements from the given white list. If the element is not one of them, revert to the default.

        Args:
            white_list (list of object): list of allowable objects
            default (object): the default fallback object
        """
        super().__init__()
        self.white_list = white_list
        self.default = default

    def to_dict(self, obj):
        if obj not in self.white_list:
            return self.default
        return obj

    def from_dict(self, value):
        if value not in self.white_list:
            return self.default
        return value



from mdt import FreeParameterTemplate, CompartmentTemplate
import numpy as np

__author__ = 'Robbert Harms'
__date__ = '2018-08-27'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class Racket(CompartmentTemplate):
    description = '''
        Computes the Racket model from [1].

        As in the original paper, we optimize "kw = k1/k2"  to preserve the inequality k1 >= k2 during optimization.

        References:
        [1] Sotiropoulos SN, Behrens TEJ, Jbabdi S. Ball and rackets: Inferring fiber fanning from 
            diffusion-weighted MRI. Neuroimage. 2012;60(2):1412-1425. doi:10.1016/j.neuroimage.2012.01.056.
    '''
    parameters = ('g', 'b', 'd', 'theta', 'phi', 'psi', 'k1', 'kw')
    dependencies = ['EigenvaluesSymmetric3x3', 'BinghamNormalization_3x3']
    cl_code = '''
        double k2 = k1 / kw;

        mot_float_type cos_theta;
        mot_float_type sin_theta = sincos(theta, &cos_theta);
        mot_float_type cos_phi;
        mot_float_type sin_phi = sincos(phi, &cos_phi);
        mot_float_type cos_psi;
        mot_float_type sin_psi = sincos(psi, &cos_psi);

        /**
         * This computes R.T*B_diag*R * -bdg^2 in one go. This code was generated using sympy and some manual adjustments:

            from sympy import symbols, Matrix, simplify, cos, sin
            psi, phi, theta, k1, k2 = symbols("psi, phi, theta, k1, k2")

            R_psi = Matrix([[cos(psi), sin(psi), 0], [-sin(psi), cos(psi), 0], [0, 0, 1]])
            R_theta = Matrix([[cos(theta), 0, -sin(theta)], [0, 1, 0], [sin(theta), 0, cos(theta)]])
            R_phi = Matrix([[cos(phi), sin(phi), 0], [-sin(phi), cos(phi), 0], [0, 0, 1]])
            B_diag = Matrix([[-k1, 0, 0], [0, -k2, 0], [0, 0, 0]])

            R = R_psi * R_theta * R_phi
            B = simplify(R.T * B_diag * R)

            Please note that B is symmetric.
        */
        mot_float_type B[9];
        B[0] = -b*d*g.x*g.x + -k1*pown(sin_phi*sin_psi - cos_phi*cos_psi*cos_theta, 2) - k2*pown(sin_phi*cos_psi + sin_psi*cos_phi*cos_theta, 2);
        B[1] = -b*d*g.x*g.y + k1*(sin_phi*sin_psi - cos_phi*cos_psi*cos_theta)*(sin_phi*cos_psi*cos_theta + sin_psi*cos_phi) - k2*(sin_phi*cos_psi + sin_psi*cos_phi*cos_theta)*(sin_phi*sin_psi*cos_theta - cos_phi*cos_psi);
        B[2] = -b*d*g.x*g.z + (-k1*(sin_phi*sin_psi - cos_phi*cos_psi*cos_theta)*cos_psi + k2*(sin_phi*cos_psi + sin_psi*cos_phi*cos_theta)*sin_psi)*sin_theta;
        B[3] = B[1];
        B[4] = -b*d*g.y*g.y + -k1*pown(sin_phi*cos_psi*cos_theta + sin_psi*cos_phi, 2) - k2*pown(sin_phi*sin_psi*cos_theta - cos_phi*cos_psi, 2);
        B[5] = -b*d*g.y*g.z + (k1*(sin_phi*cos_psi*cos_theta + sin_psi*cos_phi)*cos_psi + k2*(sin_phi*sin_psi*cos_theta - cos_phi*cos_psi)*sin_psi)*sin_theta;
        B[6] = B[2];
        B[7] = B[5];
        B[8] = -b*d*g.z*g.z + -(k1*cos_psi*cos_psi + k2*sin_psi*sin_psi)*sin_theta*sin_theta;

        mot_float_type v[3] = {k1, k2, 0};
        double denominator = BinghamNormalization_3x3(v);

        EigenvaluesSymmetric3x3(B, v);
        v[0] *= -1; v[1] *= -1; v[2] *= -1;        
        double numerator = BinghamNormalization_3x3(v);

        return (numerator/denominator);
    '''
    extra_optimization_maps = [
        lambda d: {'k2': d['k1'] / d['kw']},
        lambda d: {'ODI_k1': np.arctan2(1.0, d['k1']) * 2 / np.pi,
                   'ODI_k2': np.arctan2(1.0, d['k2']) * 2 / np.pi,
                   'ODI': np.sqrt(np.abs(np.arctan2(1.0, (d['k1'] * d['k2'])) * 2 / np.pi))}
    ]
    extra_sampling_maps = [
        lambda samples: {'k1': np.mean(samples['k1'], axis=1),
                         'k1.std': np.std(samples['k1'], axis=1),
                         'k2': np.mean(samples['k1'] / samples['kw'], axis=1),
                         'k2.std': np.std(samples['k1'] / samples['kw'], axis=1)}
    ]

    class k1(FreeParameterTemplate):
        init_value = 20
        lower_bound = 0
        upper_bound = 64
        parameter_transform = 'CosSqrClamp'
        sampling_proposal_std = 0.1
        numdiff_info = {'max_step': 0.1, 'use_upper_bound': False}

    class kw(FreeParameterTemplate):
        """We optimize the ratio w = k1/k2"""
        init_value = 4
        lower_bound = 1
        upper_bound = 64
        parameter_transform = 'CosSqrClamp'
        sampling_proposal_std = 0.1
        numdiff_info = {'max_step': 0.1, 'use_upper_bound': False}




__author__ = 'Robbert Harms'
__date__ = "2016-06-25"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ProtocolIOError(Exception):
    """Custom exception class for protocol input output errors.

    This can be raised if a protocol is inconsistent or incomplete. It should not be raised for general IO errors,
    use the IO exception for that.
    """


class InsufficientProtocolError(Exception):
    """Indicates that the protocol constains insufficient information for fitting a specific model.

    This can be raised if a model misses a column it needs in the protocol, or if there are not enough shells, etc.
    """


class NoiseStdEstimationNotPossible(Exception):
    """An exception that can be raised by any ComplexNoiseStdEstimator.

    This indicates that the noise std can not be estimated by the estimation routine.
    """


class NonUniqueComponent(Exception):
    """Raised when there are two components of the same type with the same name in the dynamically loadable components.

    If this is raised, please double check your components for items with non-unique names.
    """


class DoubleModelNameException(Exception):
    """Thrown when there are two models with the same name."""



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Create a single slice mask that only includes the voxels in the selected slice."""
import argparse
import os
import mdt
from argcomplete.completers import FilesCompleter
import textwrap
from mdt.lib.nifti import load_nifti
import mdt.utils
from mdt.lib.shell_utils import BasicShellApplication

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CreateRoiSlice(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        examples = textwrap.dedent('''
            mdt-create-roi-slice mask.nii.gz
            mdt-create-roi-slice mask.nii.gz -d 1 -s 50
            mdt-create-roi-slice mask.nii.gz -d 1 -s 50 -o my_roi_1_50.nii.gz
           ''')
        epilog = self._format_examples(doc_parser, examples)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('mask',
                            action=mdt.lib.shell_utils.get_argparse_extension_checker(['.nii', '.nii.gz', '.hdr', '.img']),
                            help='the mask to select a slice from').completer = \
            FilesCompleter(['nii', 'gz', 'hdr', 'img'], directories=False)

        parser.add_argument('-d', '--dimension', type=int, help="The dimension to index (0, 1, 2, ...). Default is 2.")
        parser.add_argument('-s', '--slice', type=int, help="The slice to use in the selected dimension (0, 1, 2, ...)."
                                                            "Defaults to center of chosen dimension.")

        parser.add_argument('-o', '--output-name',
                            action=mdt.lib.shell_utils.get_argparse_extension_checker(['.nii', '.nii.gz', '.hdr', '.img']),
                            help='the filename of the output file. Default is <mask_name>_<dim>_<slice>.nii.gz').\
            completer = FilesCompleter(['nii', 'gz', 'hdr', 'img'], directories=False)

        return parser

    def run(self, args, extra_args):
        shape = load_nifti(args.mask).shape
        roi_dimension = args.dimension if args.dimension is not None else 2
        if roi_dimension > len(shape)-1 or roi_dimension < 0:
            print('Error: the given mask has only {0} dimensions with slices {1}.'.format(len(shape), shape))
            exit(1)

        roi_slice = args.slice if args.slice is not None else shape[roi_dimension] // 2
        if roi_slice > shape[roi_dimension]-1 or roi_slice < 0:
            print('Error: dimension {0} has only {1} slices.'.format(roi_dimension, shape[roi_dimension]))
            exit(1)

        mask_base_name = os.path.splitext(os.path.realpath(args.mask))[0]
        mask_base_name = mask_base_name.replace('.nii', '')

        if args.output_name:
            output_name = os.path.realpath(args.output_name)
        else:
            output_name = mask_base_name + '_{0}_{1}.nii.gz'.format(roi_dimension, roi_slice)

        mdt.utils.write_slice_roi(os.path.realpath(args.mask), roi_dimension, roi_slice, output_name,
                                  overwrite_if_exists=True)


def get_doc_arg_parser():
    return CreateRoiSlice().get_documentation_arg_parser()


if __name__ == '__main__':
    CreateRoiSlice().start()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'TabTextual.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_TabTextual(object):
    def setupUi(self, TabTextual):
        TabTextual.setObjectName("TabTextual")
        TabTextual.resize(400, 300)
        self.gridLayout = QtWidgets.QGridLayout(TabTextual)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setHorizontalSpacing(0)
        self.gridLayout.setVerticalSpacing(4)
        self.gridLayout.setObjectName("gridLayout")
        self.textConfigEdit = TextConfigEditor(TabTextual)
        self.textConfigEdit.setObjectName("textConfigEdit")
        self.gridLayout.addWidget(self.textConfigEdit, 0, 0, 1, 1)
        self.correctness_label = QtWidgets.QLabel(TabTextual)
        self.correctness_label.setAlignment(QtCore.Qt.AlignCenter)
        self.correctness_label.setWordWrap(True)
        self.correctness_label.setObjectName("correctness_label")
        self.gridLayout.addWidget(self.correctness_label, 3, 0, 1, 1)
        self.viewSelectedOptions = QtWidgets.QCheckBox(TabTextual)
        self.viewSelectedOptions.setChecked(True)
        self.viewSelectedOptions.setObjectName("viewSelectedOptions")
        self.gridLayout.addWidget(self.viewSelectedOptions, 1, 0, 1, 1)
        self.line = QtWidgets.QFrame(TabTextual)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 2, 0, 1, 1)

        self.retranslateUi(TabTextual)
        QtCore.QMetaObject.connectSlotsByName(TabTextual)

    def retranslateUi(self, TabTextual):
        _translate = QtCore.QCoreApplication.translate
        TabTextual.setWindowTitle(_translate("TabTextual", "Form"))
        self.correctness_label.setText(_translate("TabTextual", "TextLabel"))
        self.viewSelectedOptions.setText(_translate("TabTextual", "Only show selected options"))

from ..widgets import TextConfigEditor



import logging
import matplotlib
import numpy as np
import copy
from mdt.gui.maps_visualizer.actions import SetZoom, SetAnnotations

matplotlib.use('Qt5Agg')

from PyQt5.QtCore import QTimer

from mdt.visualization.maps.matplotlib_renderer import MapsVisualizer
from matplotlib.figure import Figure
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QWidget, QVBoxLayout
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

from mdt.gui.maps_visualizer.base import DataConfigModel
from mdt.gui.maps_visualizer.renderers.base import PlottingFrame
from mdt.visualization.maps.base import Zoom, VoxelAnnotation


class MatplotlibPlotting(PlottingFrame, QWidget):

    def __init__(self, controller, parent=None, plotting_info_viewer=None):
        super().__init__(controller, plotting_info_viewer=plotting_info_viewer)

        self._controller.model_updated.connect(self.update_model)

        self._auto_render = True

        current_model = self._controller.get_model()

        self.figure = Figure(facecolor='#bfbfbf')
        self.visualizer = MapsVisualizer(current_model.get_data(), self.figure)
        self._axes_data = self.visualizer.render(current_model.get_config())

        self.canvas = FigureCanvas(self.figure)
        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.canvas.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.canvas.setFocus()
        self.canvas.updateGeometry()

        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.canvas)
        self.setLayout(layout)

        self.setParent(parent)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFocus()

        self._redraw_timer = QTimer()
        self._redraw_timer.timeout.connect(self._timer_event)
        self._redraw_timer.timeout.connect(self._redraw_timer.stop)

        self._mouse_interaction = _MouseInteraction(self.figure, self._plotting_info_viewer, self._controller)
        self._mouse_interaction.update_axes_data(self._axes_data)

        self._previous_model = None
        self.setMinimumWidth(100)

    def export_image(self, filename, width, height, dpi=100):
        current_model = self._controller.get_model()

        width_inch = width / dpi
        height_inch = height / dpi

        figure = Figure(figsize=(width_inch, height_inch), dpi=dpi)
        visualizer = MapsVisualizer(current_model.get_data(), figure)
        FigureCanvas(figure)

        visualizer.to_file(filename, current_model.get_config(), dpi=dpi)

    def set_auto_rendering(self, auto_render):
        self._auto_render = auto_render

    def redraw(self):
        self._redraw()

    @pyqtSlot()
    def _timer_event(self):
        self._redraw()

    @pyqtSlot(DataConfigModel)
    def update_model(self, model):
        def update():
            self._previous_model = model
            self.visualizer = MapsVisualizer(model.get_data(), self.figure)
            if self._auto_render:
                self._redraw_timer.start(300)

        if not self._previous_model:
            update()
        elif model.get_config().visible_changes(self._previous_model.get_config()):
            update()
        elif model.get_data() != self._previous_model.get_data():
            update()

    def _redraw(self):
        current_model = self._controller.get_model()

        self.figure.clf()

        self._axes_data = self.visualizer.render(current_model.get_config())
        self._mouse_interaction.update_axes_data(self._axes_data)

        try:
            self.figure.canvas.draw()
        except ValueError as exception:
            logger = logging.getLogger(__name__)
            logger.error(exception)


class _MouseInteraction(object):

    def __init__(self, figure, plotting_info_viewer, controller):
        self.figure = figure
        self.plotting_info_viewer = plotting_info_viewer
        self.controller = controller

        self._axes_data = []
        self.figure.canvas.mpl_connect('button_press_event', self._button_pressed)
        self.figure.canvas.mpl_connect('button_release_event', self._button_released)
        self.figure.canvas.mpl_connect('motion_notify_event', self._mouse_motion)
        self.figure.canvas.mpl_connect('scroll_event', self._scroll_event)
        self.figure.canvas.mpl_connect('key_press_event', self._on_key_press)
        self.figure.canvas.mpl_connect('key_release_event', self._on_key_release)

        self._in_drag = False
        self._control_is_held = False

        self._scrolling_manager = _ScrollingManager(controller)
        self._dragging_manager = _DraggingManager(controller)

    def update_axes_data(self, axes_data):
        """Set the updated axes data. Needs to be called if the axes are updated.

        Args:
            axes_data (list of AxisData): the information about the axes
        """
        self._axes_data = axes_data

    def _button_pressed(self, event):
        self._dragging_manager.set_starting_point(event.xdata, event.ydata)

    def _button_released(self, event):
        if self._in_drag:
            self._in_drag = False
            return

        current_annotations = copy.copy(self.controller.get_model().get_config().annotations)

        def voxel_index_in_locations(voxel_index):
            for annotation in current_annotations:
                if annotation.voxel_index == voxel_index:
                    return True
            return False

        def get_existing_matching_annotation_index(voxel_index):
            for ind, annotation in enumerate(current_annotations):
                if annotation.voxel_index == voxel_index:
                    return ind
            raise ValueError('Matching annotation not found')

        if event.button == 1:
            axis_data = self._get_matching_axis_data(event.inaxes)
            if axis_data:
                x, y = int(np.round(event.xdata)), int(np.round(event.ydata))
                index = axis_data.coordinates_to_index(x, y)[:3]

                if voxel_index_in_locations(index):
                    del current_annotations[get_existing_matching_annotation_index(index)]
                elif self._control_is_held:
                    current_annotations.append(VoxelAnnotation(index))
                else:
                    if len(current_annotations):
                        current_annotations.pop()
                    current_annotations.append(VoxelAnnotation(index))

                self.controller.apply_action(SetAnnotations(current_annotations))
            else:
                if self._control_is_held:
                    self.controller.apply_action(SetAnnotations([]))
                else:
                    if len(current_annotations):
                        current_annotations.pop()
                    self.controller.apply_action(SetAnnotations(current_annotations))

    def _scroll_event(self, event):
        if event.inaxes:
            if event.button == 'up':
                self._scrolling_manager.add_up_scroll()
            else:
                self._scrolling_manager.add_down_scroll()

    def _mouse_motion(self, event):
        if event.button == 1:
            self._in_drag = True
            self._drag_images(event)
        else:
            self._update_info_box(event)

    def _on_key_press(self, event):
        if event.key == 'control':
            self._control_is_held = True

    def _on_key_release(self, event):
        if event.key == 'control':
            self._control_is_held = False

    def _drag_images(self, event):
        self._dragging_manager.mouse_moved(event.xdata, event.ydata)

    def _update_info_box(self, event):
        """Update the info box in the plotting info object.

        Args:
            event (MouseEvent): a matplotlib mouse event used to update the plotting info
        """
        axis_data = self._get_matching_axis_data(event.inaxes)
        if axis_data:
            x, y = int(np.round(event.xdata)), int(np.round(event.ydata))
            index = axis_data.coordinates_to_index(x, y)
            self.plotting_info_viewer.set_voxel_info(axis_data.map_name, (x, y), tuple(index))
        else:
            self.plotting_info_viewer.clear_voxel_info()

    def _get_matching_axis_data(self, axis):
        """Get the axis data matching the given axis.

        Args:
            axis: the matplotlib axis to match

        Returns:
            AxisData: our data container for that axis
        """
        if axis:
            for axes_data in self._axes_data:
                if axes_data.axis == axis:
                    return axes_data
        return None


class _DraggingManager(object):

    def __init__(self, controller):
        self.controller = controller

        self._start_x = 0
        self._start_y = 0
        self._end_x = 0
        self._end_y = 0

        self._drag_timer = QTimer()
        self._drag_timer.timeout.connect(self._perform_drag)
        self._drag_timer.timeout.connect(self._drag_timer.stop)

    def set_starting_point(self, x, y):
        self._start_x = x
        self._start_y = y

    def mouse_moved(self, x, y):
        if x is not None and y is not None:
            self._end_x = x
            self._end_y = y
            self._drag_timer.start(50)

    def _perform_drag(self):
        current_model = self.controller.get_model()

        delta_x = int(np.round(self._start_x) - np.round(self._end_x))
        delta_y = int(np.round(self._start_y) - np.round(self._end_y))

        config = current_model.get_config()
        data_info = current_model.get_data()

        current_zoom = current_model.get_config().zoom

        max_y = data_info.get_max_y_index(config.dimension, rotate=config.rotate, map_names=config.maps_to_show)
        max_x = data_info.get_max_x_index(config.dimension, rotate=config.rotate, map_names=config.maps_to_show)

        new_x0 = current_zoom.p0.x + delta_x
        new_x1 = (current_zoom.p1.x or max_x) + delta_x
        new_y0 = current_zoom.p0.y + delta_y
        new_y1 = (current_zoom.p1.y or max_y) + delta_y

        if new_x0 < 0:
            new_x1 -= new_x0
            new_x0 = 0

        if new_x1 > max_x:
            new_x0 -= new_x1
            new_x1 = max_x

        if new_y0 < 0:
            new_y1 -= new_y0
            new_y0 = 0

        if new_y1 > max_y:
            new_y0 -= new_y1
            new_y1 = max_y

        try:
            new_zoom = Zoom.from_coords(new_x0, new_y0, new_x1, new_y1)
        except ValueError:
            new_zoom = current_zoom

        self.controller.apply_action(SetZoom(new_zoom))
        self._start_x = self._end_x
        self._start_y = self._end_y


class _ScrollingManager(object):

    def __init__(self, controller):
        self.controller = controller

        self._scrolls = 0

        self._scroll_timer = QTimer()
        self._scroll_timer.timeout.connect(self._perform_scroll)
        self._scroll_timer.timeout.connect(self._scroll_timer.stop)

    def add_up_scroll(self):
        self._scrolls += 1
        self._scroll_timer.start(200)

    def add_down_scroll(self):
        self._scrolls -= 1
        self._scroll_timer.start(200)

    def _perform_scroll(self):
        current_model = self.controller.get_model()
        config = current_model.get_config()
        data_info = current_model.get_data()

        current_zoom = config.zoom

        max_y = data_info.get_max_y_index(config.dimension, rotate=config.rotate, map_names=config.maps_to_show)
        max_x = data_info.get_max_x_index(config.dimension, rotate=config.rotate, map_names=config.maps_to_show)

        scrolling = self._scrolls
        if abs(self._scrolls) > 1:
            scrolling *= 3

        new_x0 = current_zoom.p0.x + scrolling
        if new_x0 < 0:
            new_x0 = 0

        new_y0 = current_zoom.p0.y + scrolling
        if new_y0 < 0:
            new_y0 = 0

        new_x1 = (current_zoom.p1.x or max_x) - scrolling
        if new_x1 > max_x:
            new_x1 = max_x

        new_y1 = (current_zoom.p1.y or max_y) - scrolling
        if new_y1 > max_y:
            new_y1 = max_y

        try:
            new_zoom = Zoom.from_coords(new_x0, new_y0, new_x1, new_y1)
        except ValueError:
            new_zoom = current_zoom

        self.controller.apply_action(SetZoom(new_zoom))
        self._scrolls = 0
